# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import integer
from .. import constants
from . import ProviderError
import arpy
from elftools.elf import elffile
from elftools.elf import relocation
from elftools.elf import sections
import struct
import sys
try:
    import builtins
except ImportError: # pragma: no cover
    import __builtin__ as builtins

class Provider(object):
    @classmethod
    def open(cls, filename_or_fileobj, filename=None):
        if hasattr(filename_or_fileobj, "read"):
            fp = filename_or_fileobj
            if filename is None:
                filename = getattr(fp, "name", None)
            needs_close = False
        else:
            assert filename is None
            filename = filename_or_fileobj
            fp = builtins.open(filename_or_fileobj, "rb")
            needs_close = True
        offset = fp.tell()
        header = fp.read(max((len(c.MAGIC) for c in cls.CLASSES)))
        for c in cls.CLASSES:
            if header.startswith(c.MAGIC):
                fp.seek(offset)
                return c(filename, fp, needs_close)
        raise ProviderError(filename, "unhandled file")

    def __init__(self, filename, fp, needs_close):
        self.filename = filename
        self.fp = fp
        self.needs_close = needs_close

    def __exit__(self, type, value, tb):
        if self.needs_close:
            self.fp.close()

class Archive(Provider):
    MAGIC = b"!<arch>\n"

    def __enter__(self):
        self.ar = arpy.Archive(self.filename, self.fp)
        self.ar.read_all_headers()
        return self

    def __exit__(self, type, value, tb):
        super(Archive, self).__exit__(type, value, tb)
        del self.ar

    @property
    def infinity_notes(self):
        for fp in self.ar.archived_files.values():
            fn = "%s[%s]" % (self.filename, fp.header.name.decode("utf-8"))
            with Provider.open(fp, fn) as fp:
                for note in fp.infinity_notes:
                    yield note

class ELF(Provider):
    MAGIC = b"\x7fELF"

    def __enter__(self):
        self.elf = elffile.ELFFile(self.fp)
        self.wordsize = self.elf.elfclass
        self.byteorder = self.elf.little_endian and b"<" or b">"
        self.__symbols = None
        self.__relocs = {}
        return self

    def __exit__(self, type, value, tb):
        super(ELF, self).__exit__(type, value, tb)
        del self.__relocs, self.__symbols, self.elf

    @property
    def infinity_notes(self):
        for sect in self.elf.iter_sections():
            if not isinstance(sect, sections.NoteSection):
                continue
            for note in sect.iter_notes():
                # pyelftools thinks n_name is "GNU", but I think
                # it should really be "GNU\0".  We'll allow both.
                if note["n_name"] not in ("GNU", "GNU\0"):
                    continue
                # Currently pyelftools returns the numeric value,
                # but at some point the constant will be added in
                # which case it's going to return a string.
                if note["n_type"] not in (constants.NT_GNU_INFINITY,
                                          "NT_GNU_INFINITY"):
                    continue
                # The inner-format part (the actual Infinity note)
                # is in the desc field.  Work out where that is in
                # relation to the containing ELF.
                elf_note_hdr_size = 12 # n_namesz, n_descsz, n_type
                elf_note_namesz_pad = ((note["n_namesz"] - 1) | 3) + 1
                i8_note_offset = (note["n_offset"]
                                  + elf_note_hdr_size
                                  + elf_note_namesz_pad)
                i8_note_bytes = note["n_desc"]
                if sys.version_info >= (3,):
                    # pyelftools converts to string, but we want bytes
                    i8_note_bytes = i8_note_bytes.encode("latin-1")
                assert len(i8_note_bytes) == note["n_descsz"]
                yield NoteSlice(Note(self, sect, i8_note_offset,
                                     i8_note_bytes),
                                slice(0, note["n_descsz"]))

    @property
    def symbols(self):
        if self.__symbols is None:
            self.__symbols = {}
            for sect in self.elf.iter_sections():
                if not isinstance(sect, sections.SymbolTableSection):
                    continue
                for sym in sect.iter_symbols():
                    addr = sym["st_value"]
                    if addr not in self.__symbols:
                        self.__symbols[addr] = [sym.name]
                    else:
                        self.__symbols[addr].append(sym.name)
        return self.__symbols

    def __get_named_symbol(self, symtab, index):
        symbol = symtab.get_symbol(index)
        if symbol.name:
            return symbol
        # XXX this symbol is the start of a section?
        assert symbol.entry["st_info"]["type"] == "STT_SECTION"
        assert symbol.entry["st_value"] == 0
        st_shndx = symbol.entry["st_shndx"]
        for symbol in symtab.iter_symbols():
            if (symbol.entry["st_shndx"] == st_shndx
                    and symbol.entry["st_value"] == 0
                    and symbol.name):
                return symbol
        raise ProviderError(self.filename, "unhandled relocation")

    def relocations_for(self, sect):
        relocs = self.__relocs.get(sect.name, None)
        if relocs is None:
            relocs = self.__relocs[sect.name] = {}
            reloc_handler = relocation.RelocationHandler(self.elf)
            reloc_sect = reloc_handler.find_relocations_for_section(sect)
            symtab = self.elf.get_section(reloc_sect["sh_link"])
            for reloc in reloc_sect.iter_relocations():
                symbol = self.__get_named_symbol(symtab, reloc["r_info_sym"])
                relocs[reloc["r_offset"]] = symbol.name
        return relocs

    def symbol_names_at(self, section, offset, addr):
        if addr == 0:
            return [self.relocations_for(section)[offset]]
        else:
            return self.symbols[addr]

Provider.CLASSES = [Archive, ELF]
open = Provider.open

class Note(object):
    def __init__(self, elf, section, offset, bytes):
        self.elf = elf
        self.section = section
        self.offset = offset # of first byte of n_desc in ELF
        self.bytes = bytes

    @property
    def filename(self):
        return self.elf.filename

    @property
    def wordsize(self):
        return self.elf.wordsize

    @property
    def byteorder(self):
        return self.elf.byteorder

    def symbol_names_at(self, offset):
        fmt = self.byteorder + {32: b"I", 64: b"Q"}[self.wordsize]
        size = struct.calcsize(fmt)
        addr = struct.unpack(fmt, self.bytes[offset:offset + size])[0]
        offset = self.offset + offset - self.section["sh_offset"]
        return self.elf.symbol_names_at(self.section, offset, addr)

class NoteSlice(object):
    def __init__(self, note, key):
        assert isinstance(key, slice)
        assert key.step in (None, 1)
        self.note = note

        assert key.start >= 0
        self.start = key.start
        assert key.stop <= len(note.bytes)
        self.limit = key.stop

    def __len__(self):
        return self.limit - self.start

    def __getitem__(self, key):
        if isinstance(key, integer):
            result = self.bytes[key]
            if isinstance(result, int):
                result = chr(result)
            return result

        assert isinstance(key, slice)
        assert key.step in (None, 1)

        start = self.start
        if key.start is not None:
            start += key.start
            assert start <= self.limit
        if key.stop is not None:
            limit = self.start + key.stop
            assert limit <= self.limit
        else:
            limit = self.limit
        return NoteSlice(self.note, slice(start, limit))

    def __add__(self, offset):
        start = self.start + offset
        assert start <= self.limit
        return NoteSlice(self.note, slice(start, self.limit))

    @property
    def filename(self):
        return self.note.filename

    @property
    def wordsize(self):
        return self.note.wordsize

    @property
    def byteorder(self):
        return self.note.byteorder

    @property
    def bytes(self):
        return self.note.bytes[self.start:self.limit]

    @property
    def text(self):
        text = self.bytes
        if sys.version_info >= (3,):
            text = "".join(map(chr, text))
        assert isinstance(text, str)
        return text

    @property
    def symbol_names(self):
        return self.note.symbol_names_at(self.start)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

# Operators which can have folded loads

comparisons = ("lt", "le", "eq", "ne", "ge", "gt")

BINARY_OPS = ("add", "and", "div", "mod", "mul",
              "shl", "shr", "shra", "sub", "xor") + comparisons

BRANCH_OPS = tuple("b" + op for op in comparisons)

# Sources for the various tests

BINARY_SOURCE = """\
define test::fold_load_test returns int
    argument int x
    %s
"""

BRANCH_SOURCE = BINARY_SOURCE + """\
    load 0
    return
label:
    load 1
    return
"""

CALL_SOURCE = """\
define test::fold_load_test
    %s
"""

DEREF_SOURCE = """\
define test::fold_load_test returns ptr
    extern ptr p
    %s
"""

class TestFoldLoad(TestCase):
    def test_binary_fold_load(self):
        """Check that folded loads work for binary operations."""
        self.__do_test(BINARY_SOURCE, BINARY_OPS, "5")

    def test_branch_fold_load(self):
        """Check that folded loads work for conditional branches."""
        self.__do_test(BRANCH_SOURCE, BRANCH_OPS, "5", "label")

    def test_call_fold_load(self):
        """Check that folded loads work for "call"."""
        self.__do_test(CALL_SOURCE, ("call",), "fold_load_test")

    def test_deref_fold_load(self):
        """Check that folded loads work for "deref"."""
        self.__do_test(DEREF_SOURCE, ("deref",), "p", "ptr")

    def __do_test(self, source, ops, load_arg, extra=None):
        for op in ops:
            expect_ops = [{"call": "load_external",
                           "deref": "addr"}.get(op, "lit%s" % load_arg),
                          {"add": "plus",
                           "sub": "minus"}.get(op, op)]
            if expect_ops[1] == "plus":
                expect_ops = ["plus_uconst"]
            elif expect_ops[1] in BRANCH_OPS:
                expect_ops[1] = expect_ops[1][1:]

            sub1 = "load %s\n    %s" % (load_arg, op)
            sub2 = "%s %s" % (op, load_arg)
            if extra is not None:
                sub1 += " " + extra
                sub2 += ", " + extra

            for sub in sub1, sub2:
                tree, output = self.compile(source % sub)
                self.assertEqual(expect_ops, output.opnames)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import cmdline
from .. import version
from ..compat import fprint, load_module_from_source, strtoint_c
from . import context
from . import I8XError
from . import TestCase
import getopt
import os
import sys
try:
    import unittest2 as unittest
except ImportError: # pragma: no cover
    import unittest

USAGE = """\
Usage: i8x [OPTION]... TESTFILE...
   or: i8x [OPTION]... [-q|--quick] FUNCTION [ARGUMENT]...

Infinity Note Execution Environment.

Options:
  --help                Display this information.
  --version             Display version information.
  -I DIR                Add the directory DIR to the list of directories
                        that TestCase.import_constants_from will search
                        for header files.
  -i, --import=ELFFILE  Import notes from ELFFILE.
  -q, --quick           Execute the function and arguments specified on
                        command line.
  -t, --trace           Trace function execution.  This option may be
                        specified multiple times for greater detail.""" \
    + cmdline.usage_message_footer_for("I8X")

LICENSE = ("LGPLv2.1+: GNU LGPL version 2.1 or later",
           "http://www.gnu.org/licenses/lgpl-2.1.html")

def module_install_command(module):
    dir = os.path.dirname(os.path.realpath(sys.executable))
    for cmd in ("pip", "pip3"):
        cmd = os.path.join(dir, cmd)
        if os.path.exists(cmd):
            return "%s install %s" % (cmd, module)
    for cmd in ("easy_install",):
        cmd = os.path.join(dir, cmd)
        if os.path.exists(cmd):
            return "%s %s" % (cmd, module)

class TestSuite(unittest.TestSuite):
    def __init__(self, *args, **kwargs):
        unittest.TestSuite.__init__(self, *args, **kwargs)
        self.__loader = unittest.TestLoader()

    def load_i8tests(self, ctx, filename):
        name = os.path.splitext(os.path.basename(filename))[0]
        module = load_module_from_source(name, filename)
        for name in dir(module):
            item = getattr(module, name)
            if (item is not TestCase
                and type(item) is type
                and issubclass(item, TestCase)):
                self.addTest(self.__loader.loadTestsFromTestCase(item))

def main(args):
    clue = "Try ‘i8x --help’ for more information."
    try:
        opts, args = getopt.gnu_getopt(
            args,
            "i:I:qt",
            ("help", "version", "import=", "quick", "trace"))
    except getopt.GetoptError as e:
        raise I8XError("%s\n%s" % (e, clue))
    ctx = context.Context()
    quickmode = False
    for opt, arg in opts:
        if opt == "--help":
            fprint(sys.stdout, USAGE)
            return
        elif opt == "--version":
            fprint(sys.stdout, cmdline.version_message_for("I8X", LICENSE))
            return
        elif opt == "-I":
            TestCase.include_path.append(arg)
        elif opt in ("-i", "--import"):
            ctx.import_notes(arg)
        elif opt in ("-q", "--quick"):
            quickmode = True
        elif opt in ("-t", "--trace"):
            ctx.tracelevel += 1

    if len(args) < 1:
        raise I8XError("nothing to do!\n%s" % clue)

    if quickmode:
        function = args.pop(0)
        args = [strtoint_c(arg, I8XError) for arg in args]
        result = map(str, ctx.call(function, *args))
        fprint(sys.stdout, ", ".join(result))
        return

    if not hasattr(TestCase, "assertIsInstance"):
        msg = ("unittest2 is required to run testcases"
               + " with Python %s.%s" % sys.version_info[:2])
        clue = module_install_command("unittest2")
        if clue is not None:
            msg += "\nTry ‘%s’" % clue
        raise I8XError(msg)

    print("I8X", version(), "on Python", sys.version)
    print()

    TestCase.i8ctx = ctx

    tests = TestSuite()
    for filename in args:
        tests.load_i8tests(ctx, filename)

    result = unittest.TextTestRunner(stream=sys.stdout,
                                     verbosity=2).run(tests)
    if not result.wasSuccessful():
        return 1

    report = []
    ops_hit = opcount = maxlen = 0
    for sig, funclist in sorted(ctx.functions.items()):
        for function in funclist:
            hit, count = function.coverage
            ops_hit += hit
            opcount += count
            maxlen = max(len(sig), maxlen)
            report.append((sig, hit, count))

    if ops_hit != opcount:
        if hasattr(result, "separator2"):
            print(result.separator2)
        print("Coverage:")
        print()

        rowfmt = "%%-%ds %%3.0f%%%%" % maxlen
        for sig, hit, count in report:
            print(rowfmt % (sig, 100 * hit / count))



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from i8c import I8Error

class I8XError(I8Error):
    """Base class for all runtime errors.
    """
    def __init__(self, msg, prefix="i8x"):
        I8Error.__init__(self, msg, prefix)

class InputFileError(I8XError):
    """An error occurred processing one of our input files.
    """
    def __init__(self, filename, msg):
        I8XError.__init__(self, msg, filename)

class ProviderError(InputFileError):
    """An error occurred processing a note-providing file.
    """

class TestFileError(InputFileError):
    """An error occurred processing a Python test file.
    """

class HeaderFileError(InputFileError):
    """An error occurred processing a C header file.
    """
    def __init__(self, filename, linenumber):
        I8XError.__init__(self,
                          "expected ‘#define NAME VALUE’",
                          "%s:%d" % (filename, linenumber))

class NoteError(I8XError):
    """An error was detected while decoding a note.
    """
    def __init__(self, es, msg):
        I8XError.__init__(self, msg, "%s[0x%08x]" % (es.filename, es.start))

class CorruptNoteError(NoteError):
    """A corrupt note was detected.
    """
    def __init__(self, elfslice):
        NoteError.__init__(self, elfslice, "corrupt note")

class UnhandledNoteError(NoteError):
    """An unhandled note was detected.
    """
    def __init__(self, elfslice):
        NoteError.__init__(self, elfslice, "unhandled note")

class FunctionLookupError(I8XError):
    """An error occurred during function lookup.
    """
    def __init__(self, msg, reference=None):
        if reference is None:
            I8XError.__init__(self, msg)
        else:
            I8XError.__init__(self, msg, reference.referrer)

class UndefinedFunctionError(FunctionLookupError):
    """The requested function was not defined.
    """
    def __init__(self, signature, reference=None):
        FunctionLookupError.__init__(
            self, "undefined function ‘%s’" % signature, reference)

class AmbiguousFunctionError(FunctionLookupError):
    """Multiple copies of the requested function were found.
    """
    def __init__(self, signature, reference=None):
        FunctionLookupError.__init__(
            self, "function ‘%s’ defined more than once" % signature,
            reference)

class ExecutionError(I8XError):
    """An error was detected during bytecode execution.
    """
    def __init__(self, op, msg):
        I8XError.__init__(self, msg, "%s+%04d" % op.location)

class BadJumpError(ExecutionError):
    """A DW_OP_bra or DW_OP_skip went to a bad place.
    """
    def __init__(self, op):
        ExecutionError.__init__(self, op, "bad jump")

class BadDerefError(ExecutionError):
    """Something dereferenced an invalid location.
    """
    def __init__(self, op, mem, loc):
        ExecutionError.__init__(
            self, op, "0x%08x: invalid location:\n%s" % (loc, mem))



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import next
from . import lexer
from . import logger
from . import ParserError
from . import ParsedError
from . import visitors
import copy
from operator import eq as EXACTLY, ge as ATLEAST

debug_print = logger.debug_printer_for(__name__)

def raise_unless_len(tokens, cmp, count):
    if cmp(len(tokens), count):
        return
    # Try to narrow down to an offending token
    if cmp == EXACTLY and len(tokens) > count:
        index = count
    else:
        index = -1
    raise ParserError(tokens[index:])

class TreeNode(visitors.Visitable):
    def __init__(self):
        self.tokens = []
        self.children = []

    @property
    def fileline(self):
        return self.tokens[0].fileline

    def add_child(self, klass):
        child = klass()
        self.children.append(child)
        return child

    @property
    def latest_child(self):
        return self.children[-1]

    def some_children(self, classinfo):
        return (child
                for child in self.children
                if isinstance(child, classinfo))

    def one_child(self, classinfo):
        result = None
        for child in self.some_children(classinfo):
            if result is not None:
                self.__one_child_error("multiple", classinfo)
            result = child
        if result is None:
            self.__one_child_error("no", classinfo)
        return result

    def __one_child_error(self, what, classinfo):
        msg = "%s has %s %s" % (
            self.__class__.__name__,
            what,
            classinfo.__name__)
        raise ParsedError(self, msg.lower())

    def __str__(self):
        lines = []
        self.__dump(lines, "")
        return "\n".join(lines)

    def __dump(self, lines, prefix):
        line = prefix + self.__class__.__name__
        if isinstance(self, LeafNode):
            line += ': '
            line += " ".join((token.text for token in self.tokens))
        elif isinstance(self, Operation) and not self.has_own_handler:
            line += '("%s")' % self.tokens[0].text
        if hasattr(self, "type"):
            # Dump what the type annotator added
            line += " [%s" % self.type.name
            basetype = self.type.basetype
            if basetype is not self.type:
                line += " = %s" % basetype.name
            line += "]"
        if (isinstance(self, (FullName, ShortName))
            and hasattr(self, "value")):
            # Dump what the name annotator added
            line += " [%s]" % self.value
        lines.append(line)
        for child in self.children:
            child.__dump(lines, prefix + "  ")

class LeafNode(TreeNode):
    def consume(self, tokens):
        if self.tokens:
            raise ParserError(tokens)
        self.tokens = tokens

class SyntheticNode(LeafNode):
    """A node that the block creator created.
    """
    def __init__(self, template, text):
        assert isinstance(template, TreeNode)
        self.tokens = [lexer.SyntheticToken(template.tokens[0],
                                            "synthetic " + text)]

class Identifier(LeafNode):
    def consume(self, tokens):
        raise_unless_len(tokens, EXACTLY, 1)
        if not isinstance(tokens[0], lexer.WORD):
            raise ParserError(tokens)
        LeafNode.consume(self, tokens)

    @property
    def name(self):
        return self.tokens[0].text

class Constant(LeafNode):
    def consume(self, tokens):
        raise_unless_len(tokens, EXACTLY, 1)
        LeafNode.consume(self, tokens)

class Integer(Constant):
    def consume(self, tokens):
        Constant.consume(self, tokens)
        if not isinstance(self.tokens[0], lexer.NUMBER):
            raise ParserError(tokens)
        self.value = self.tokens[0].value

class String(Constant):
    def consume(self, tokens):
        Constant.consume(self, tokens)
        if not isinstance(self.tokens[0], lexer.STRING):
            raise ParserError(tokens)
        self.value = self.tokens[0].value

class BuiltinConstant(Constant):
    def consume(self, tokens):
        Constant.consume(self, tokens)
        self.value = self.VALUES.get(tokens[0].text, None)
        if self.value is None:
            raise ParserError(tokens)

class Pointer(BuiltinConstant):
    VALUES = {"NULL": 0}

class Boolean(BuiltinConstant):
    VALUES = {"TRUE": 1, "FALSE": 0}

class TopLevel(TreeNode):
    def consume(self, tokens):
        klass = self.CLASSES.get(tokens[0].text, None)
        if klass is Wordsize and self.children:
            raise ParserError(tokens)
        if klass is not None:
            if (klass is not External
                  or not self.children
                  or not isinstance(self.latest_child, Function)):
                self.add_child(klass)
        elif not self.children:
            raise ParserError(tokens)
        self.latest_child.consume(tokens)

    @property
    def wordsize_directives(self):
        return self.some_children(Wordsize)

    @property
    def typedefs(self):
        return self.some_children(Typedef)

    @property
    def functions(self):
        return self.some_children(Function)

class Typedef(TreeNode):
    def consume(self, tokens):
        raise_unless_len(tokens, ATLEAST, 3)
        if self.tokens:
            raise ParserError(tokens)
        self.tokens = tokens
        self.add_child(TypeName).consume([tokens[-1]])
        tokens = tokens[1:-1]
        self.add_child(Type.class_for(tokens)).pop_consume(tokens)

class TypeName(Identifier):
    pass

class Type:
    @staticmethod
    def class_for(tokens):
        if tokens[0].text == "func":
            return FuncType
        else:
            return BasicType

class BasicType(Identifier, Type):
    def pop_consume(self, tokens):
        if self.tokens:
            raise ParserError(tokens)
        self.tokens = [tokens.pop(0)]

class FuncType(TreeNode, Type):
    def pop_consume(self, tokens):
        if self.tokens:
            raise ParserError(tokens)
        self.tokens = [tokens.pop(0)]
        self.add_child(ReturnTypes).pop_consume(tokens, lexer.OPAREN)
        assert tokens and isinstance(tokens[0], lexer.OPAREN)
        tokens.pop(0)
        self.add_child(ParamTypes).pop_consume(tokens, lexer.CPAREN)
        assert tokens and isinstance(tokens[0], lexer.CPAREN)
        tokens.pop(0)

    @property
    def returntypes(self):
        return self.one_child(ReturnTypes)

    @property
    def paramtypes(self):
        return self.one_child(ParamTypes)

class TypeList(TreeNode):
    def pop_consume(self, tokens, stop_at=None):
        self.tokens = copy.copy(tokens)
        while tokens:
            if stop_at is not None and isinstance(tokens[0], stop_at):
                break
            self.add_child(Type.class_for(tokens)).pop_consume(tokens)
            if tokens and isinstance(tokens[0], lexer.WORD):
                # This is a named type.  We've no use for the name
                # right now, but we could replace self.children[-1]
                # with a TypeAndName if we need it for something.
                tokens.pop(0)
            if tokens:
                if stop_at is not None and isinstance(tokens[0], stop_at):
                    break
                if not isinstance(tokens[0], lexer.COMMA):
                    raise ParserError(tokens)
                tokens.pop(0)
        else:
            if stop_at is not None:
                raise ParserError(tokens)

class ReturnTypes(TypeList):
    pass

class ParamTypes(TypeList):
    pass

class Function(TreeNode):
    def consume(self, tokens):
        if not self.tokens:
            raise_unless_len(tokens, ATLEAST, 4)
            self.tokens = tokens
            self.add_child(FullName).consume(tokens[1:4])
            tokens = tokens[4:]
            if tokens:
                if tokens[0].text != "returns":
                    raise ParserError(tokens)
                tokens.pop(0)
            self.returntypes = self.add_child(ReturnTypes)
            self.returntypes.pop_consume(tokens)
            self.parameters = self.add_child(Parameters)
            self.externals = self.add_child(Externals)
            self.ops = self.add_child(Operations)
            return

        if tokens[0].text == "argument":
            if self.ops.children:
                raise ParserError(tokens)
            self.parameters.consume(tokens)
        elif tokens[0].text == "extern":
            if self.ops.children:
                raise ParserError(tokens)
            self.externals.consume(tokens)
        else:
            self.ops.consume(tokens)

    @property
    def name(self):
        return self.one_child(FullName)

    @property
    def provider(self):
        return self.name.provider

class FullName(TreeNode):
    def consume(self, tokens):
        raise_unless_len(tokens, EXACTLY, 3)
        if (self.tokens
            or not isinstance(tokens[1], lexer.DOUBLE_COLON)):
            raise ParserError(tokens)
        self.tokens = tokens
        self.add_child(Provider).consume([tokens[0]])
        self.add_child(ShortName).consume([tokens[2]])

    @property
    def provider(self):
        return self.one_child(Provider).name

    @property
    def shortname(self):
        return self.one_child(ShortName).name

    @property
    def ident(self):
        """How will the user refer to this item in the source?"""
        return "%s::%s" % (self.one_child(Provider).name,
                           self.one_child(ShortName).name)

class Provider(Identifier):
    pass

class ShortName(Identifier):
    pass

class TypeAndName(TreeNode):
    def consume(self, tokens):
        self.add_child(Type.class_for(tokens)).pop_consume(tokens)
        if self.allow_fullname and len(tokens) == 3:
            klass = FullName
        elif len(tokens) == 1:
            klass = ShortName
        else:
            raise ParserError(self.tokens)
        self.add_child(klass).consume(tokens)

    @property
    def typename(self):
        return self.one_child(Type)

    @property
    def name(self):
        return self.one_child((ShortName, FullName))

class Parameters(TreeNode):
    def consume(self, tokens):
        if not self.tokens:
            self.tokens = tokens
        self.add_child(Parameter).consume(tokens)

class Parameter(TypeAndName):
    allow_fullname = False

    def consume(self, tokens):
        if self.tokens:
            raise ParserError(tokens)
        self.tokens = tokens
        TypeAndName.consume(self, tokens[1:])

class Externals(TreeNode):
    def consume(self, tokens):
        if not self.tokens:
            self.tokens = tokens
        self.add_child(External).consume(tokens)

class External(TypeAndName):
    allow_fullname = True

    def consume(self, tokens):
        if self.tokens:
            raise ParserError(tokens)
        self.tokens = tokens
        TypeAndName.consume(self, tokens[1:])

class Wordsize(TreeNode):
    def consume(self, tokens):
        raise_unless_len(tokens, EXACTLY, 2)
        if self.tokens:
            raise ParserError(tokens)
        self.add_child(Integer).consume(tokens[1:])

TopLevel.CLASSES = {
    "define": Function,
    "extern": External,
    "typedef": Typedef,
    "wordsize": Wordsize}

# XXX

class Label(Identifier):
    def consume(self, tokens):
        Identifier.consume(self, [tokens[0]])

# Base class for all operations

class Operation(TreeNode):
    has_own_handler = False
    may_fold_load = False

    def consume(self, tokens):
        if self.tokens:
            raise ParserError(tokens)
        self.tokens = tokens

        # Split the argument list
        tokens, args = tokens[1:], []
        while tokens:
            if isinstance(tokens[0], lexer.COMMA):
                # Comma immediately after operator,
                # or comma immediately after comma
                raise ParserError(tokens)
            arg = []
            while tokens:
                token = tokens.pop(0)
                if not isinstance(token, lexer.COMMA):
                    arg.append(token)
                elif not tokens:
                    # Comma at end of line
                    raise ParserError(token)
                else:
                    break
            assert arg
            args.append(arg)

        # Handle any folded loads
        if self.may_fold_load and len(args) == self.num_args + 1:
            self.add_folded_load(args.pop(0))

        # Process the arguments
        if len(args) > self.num_args:
            raise ParserError(args[self.num_args])
        elif len(args) < self.num_args:
            raise ParserError(self.tokens[-1:])
        else:
            self.add_children(*args)

    def add_folded_load(self, tokens):
        token = lexer.SyntheticToken(tokens[0], "synthetic load")
        self.add_child(LoadOp).consume([token] + tokens)

    @property
    def folded_children(self):
        return self.some_children(Operation)

    @property
    def operand(self):
        non_folded = [child
                      for child in self.children
                      if not isinstance(child, Operation)]
        assert len(non_folded) == 1
        return non_folded[0]

# XXX blah blah blah

class NoArgOp(Operation):
    num_args = 0

    def add_children(self):
        pass

class OneArgOp(Operation):
    num_args = 1

class TwoArgOp(Operation):
    num_args = 2

class JumpOp(OneArgOp):
    def add_children(self, target):
        self.add_child(Target).consume(target)

    @property
    def target(self):
        return self.one_child(Target)

class NameCastOp(TwoArgOp):
    has_own_handler = True

    def add_children(self, slot, arg2):
        if isinstance(slot[0], lexer.NUMBER):
            slotclass = StackSlot
        else:
            slotclass = ShortName
        self.slot = self.add_child(slotclass)
        self.slot.consume(slot)
        self.add_child_2(arg2)

    @property
    def named_operands(self):
        """Operands that need processing by the name annotator.
        """
        return self.some_children(ShortName)

class Target(Identifier):
    pass

class StackSlot(Integer):
    def consume(self, tokens):
        Integer.consume(self, tokens)
        if self.value < 0:
            raise ParserError(tokens)

# Classes which represent groups of operators in the parse tree

class SimpleOp(NoArgOp):
    @property
    def name(self):
        return self.tokens[0].text

class SimpleFoldLoadOp(SimpleOp):
    may_fold_load = True

class CompareOp(NoArgOp):
    may_fold_load = True

class CondBranchOp(JumpOp):
    may_fold_load = True

# Classes for operators that require specific individual parsing

class AddOp(SimpleFoldLoadOp):
    has_own_handler = True
    name = "add"

class CastOp(NameCastOp):
    def add_child_2(self, type):
        raise_unless_len(type, EXACTLY, 1)
        self.add_child(BasicType).pop_consume(type)

    @property
    def typename(self):
        return self.one_child(BasicType)

class DerefOp(OneArgOp):
    has_own_handler = True
    may_fold_load = True

    def add_folded_load(self, tokens):
        if (len(tokens) == 4
              and isinstance(tokens[1], lexer.OPAREN)
              and isinstance(tokens[3], lexer.CPAREN)):
            # Handle "offset(base)" syntax
            OneArgOp.add_folded_load(self, [tokens[2]])
            OneArgOp.add_folded_load(self, [tokens[0]])
            token = lexer.SyntheticToken(tokens[0], "synthetic add")
            self.add_child(AddOp).consume([token])
        else:
            # Regular folded load
            OneArgOp.add_folded_load(self, tokens)

    def add_children(self, type):
        raise_unless_len(type, EXACTLY, 1)
        self.add_child(BasicType).pop_consume(type)

class DupOp(NoArgOp):
    has_own_handler = True

class GotoOp(JumpOp):
    has_own_handler = True

class LoadOp(OneArgOp):
    has_own_handler = True

    def add_children(self, arg):
        if len(arg) == 1:
            if isinstance(arg[0], lexer.NUMBER):
                klass = Integer
            else:
                for klass in Pointer, Boolean:
                    if arg[0].text in klass.VALUES:
                        break
                else:
                    klass = ShortName
        elif len(arg) == 3:
            klass = FullName
        else:
            raise ParserError(arg)
        self.add_child(klass).consume(arg)

    @property
    def is_offset_base(self):
        return len(list(self.folded_children)) == 2

    @property
    def name(self):
        return self.one_child((ShortName, FullName))

    @property
    def named_operands(self):
        """Operands that need processing by the name annotator.
        """
        return self.some_children((ShortName, FullName))

    @property
    def typed_operands(self):
        """Operands that need processing by the type annotator.
        """
        return self.some_children(Constant)

class NameOp(NameCastOp):
    def add_child_2(self, newname):
        self.newname = self.add_child(ShortName)
        self.newname.consume(newname)

class OverOp(NoArgOp):
    has_own_handler = True

class PickOp(OneArgOp):
    has_own_handler = True

    def add_children(self, slot):
        self.add_child(StackSlot).consume(slot)

class ReturnOp(NoArgOp):
    has_own_handler = True

class WarnOp(OneArgOp):
    has_own_handler = True

    def add_children(self, text):
        self.add_child(String).consume(text)

# XXX

class Operations(TreeNode):
    CLASSES = {"add": AddOp,
               "cast": CastOp,
               "deref": DerefOp,
               "dup": DupOp,
               "goto": GotoOp,
               "load": LoadOp,
               "name": NameOp,
               "over": OverOp,
               "pick": PickOp,
               "return": ReturnOp,
               "warn": WarnOp}
    for op in ("abs", "drop", "neg", "not", "rot", "swap"):
        CLASSES[op] = SimpleOp
    for op in ("and", "div", "call", "mod", "mul", "or",
               "shl", "shr", "shra", "sub", "xor"):
        CLASSES[op] = SimpleFoldLoadOp
    for op in ("lt", "le", "eq", "ne", "ge", "gt"):
        CLASSES[op] = CompareOp
        CLASSES["b" + op] = CondBranchOp
    del op

    # Do not add an "addr" instruction for DW_OP_addr.  Users
    # should define an "extern ptr" and then "load" it.
    assert "addr" not in CLASSES

    # Do not add a "bra" instruction, it gives no clue as to why
    # it would branch and makes code harder to read.  Users should
    # use "bne NULL" or "bne 0" and let the optimizer figure it out.
    assert "bra" not in CLASSES

    # Do not add "plus" or "minus" operations.  To calculate the
    # sum of two values you add them.  To calculate the difference
    # of two values you sub(tract) them.  That is all.
    assert "plus" not in CLASSES
    assert "minus" not in CLASSES

    def consume(self, tokens):
        if not self.tokens:
            self.tokens = tokens
        if len(tokens) == 2 and isinstance(tokens[1], lexer.COLON):
            klass = Label
        else:
            klass = self.CLASSES.get(tokens[0].text, None)
            if klass is None:
                raise ParserError(tokens)
        self.add_child(klass).consume(tokens)

def build_tree(tokens):
    tree = TopLevel()
    try:
        group = None
        terminators = []
        while True:
            if group is None:
                group = []
                terminators.append(lexer.NEWLINE)
            try:
                token = next(tokens)
            except StopIteration:
                break
            if isinstance(token, terminators[-1]):
                terminators.pop()
                if not terminators:
                    tree.consume(group)
                    group = None
                    continue
            elif isinstance(token, lexer.OPAREN):
                terminators.append(lexer.CPAREN)
            if not isinstance(token, lexer.NEWLINE):
                group.append(token)
        if group:
            if terminators == [lexer.NEWLINE]:
                # No newline at end of input
                tree.consume(group)
            else:
                # Unclosed parenthesis?
                raise NotImplementedError
        return tree
    finally:
        if debug_print.is_enabled:
            debug_print("%s\n\n" % tree)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

# The final "call" was dropped as unreachable.

SOURCE = """\
define test::i8cfail_0001 returns int, ptr
    argument int arg1
    extern func int, ptr (int) func1
    extern func int () func2
    extern ptr sym1

    load func1
    load func2
    load sym1

    deref ptr
    load NULL
    bne label2
    swap
    drop
    load NULL
    rot
    call
    beq label1
    load 1
    return
label1:
    load 0
    return
label2:
    drop
    call
"""

class TestI8CFail0001(TestCase):
    def test_i8cfail_0001(self):
        """Miscellaneous I8C failure #0001 check"""
        tree, output = self.compile(SOURCE)
        ops = output.opnames
        # Robust test: check the stream contains two calls. They
        # refer to different functions so there's no way the two
        # can be combined.
        self.assertEqual(len([op for op in ops if op == "call"]), 2)
        # Fragile test: check that the stream is as we expect.
        # This may need fixing up to cope with future compiler changes.
        self.assertEqual(["load_external", "load_external",
                          "addr",
                          "deref", "bra", "swap", "drop",
                          "lit0", "rot", "call", "ne",
                          "skip", "drop", "call"], ops)



# -*- coding: utf-8 -*-
# Copyright (C) 2016 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

SOURCE = """\
define test::test_warn returns ptr
    argument ptr arg
    dup
    bne NULL, label
    warn "hello world"
    warn ""
    warn "it's \\"NULL\\""
label:
"""

class TestWarn(TestCase):
    """Check warnings."""

    def setUp(self):
        self.warnings = []

    def warn_caller(self, msg):
        self.warnings.append(msg)

    def test_warning(self):
        """Basic check of a warning."""
        tree, output = self.compile(SOURCE)
        output.call("test::test_warn(p)p", 5)
        self.assertEqual(self.warnings, [])
        output.call("test::test_warn(p)p", 0)
        self.assertEqual(self.warnings, ["hello world",
                                         "",
                                         'it\'s "NULL"'])



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import StackError

SOURCE = """\
define test::binary_ops_test
    argument %s arg_1
    argument %s arg_2
    %s
"""

class TestBinaryOperations(TestCase):
    OPERATIONS = ("add", "and", "div", "mod", "mul", "or",
                  "shl", "shr", "shra", "sub", "xor")
    TYPES = "int", "bool", "ptr", "opaque", "func ()"

    def test_binary_ops(self):
        """Check that binary operations work."""
        for op in self.OPERATIONS:
            for type1 in self.TYPES:
                for type2 in self.TYPES:
                    self.__run_test(op, type1, type2)

    def __run_test(self, op, type1, type2):
        """Try and compile TYPE1 OP TYPE2."""
        source = SOURCE % (type1, type2, op)
        # NB Top of stack is TYPE2, second entry is TYPE1.
        if self.__expect_success(op, type1, type2):
            tree, output = self.compile(source)
            op = {"add": "plus", "sub": "minus"}.get(op, op)
            self.assertEqual([op], output.opnames)
        else:
            self.assertRaises(StackError, self.compile, source)

    def __expect_success(self, op, type1, type2):
        """Is TYPE1 OP TYPE2 a valid thing to do?"""
        t1int = type1 in ("int", "bool")
        t2int = type2 in ("int", "bool")
        if t1int and t2int:
            return True
        if (op == "add"
            and (t1int or t2int)
            and "ptr" in (type1, type2)):
            return True
        if (op == "sub"
            and t2int
            and type1 == "ptr"):
            return True
        return False



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

class VisitError(NotImplementedError):
    def __init__(self, visitor, visitee):
        NotImplementedError.__init__(
            self,
            "%s has no visitor suitable for %s" % (
                    visitor.__class__.__name__,
                    visitee.__class__.__name__))

class Visitable(object):
    @classmethod
    def get_visitfunc(cls, visitor):
        func = getattr(visitor, "visit_" + cls.__name__.lower(), None)
        if func is None:
            for cls in cls.__bases__:
                if cls is not Visitable and issubclass(cls, Visitable):
                    func = cls.get_visitfunc(visitor)
                    if func is not None:
                        break
        return func

    def accept(self, visitor):
        # Find a suitable visitor
        func = self.get_visitfunc(visitor)
        if func is None:
            raise VisitError(visitor, self)
        # Visit any folded children first
        if hasattr(self, "folded_children"):
            for node in self.folded_children:
                node.accept(visitor)
        # Now visit ourself
        return func(self)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c import compiler
from i8c import runtime
import os
import sys

class TestExamples(TestCase):
    def setUp(self):
        # Locate the directories
        self.__examples = os.path.join(self.topdir, "examples")
        self.__output = os.path.join(self.topdir, "tests",
                                     "output", "test_examples")
        if not os.path.exists(self.__output):
            os.makedirs(self.__output)
        # Create a list of files to delete
        self.__unlink_me = []
        # Save whatever sys.argv is
        self.__saved_argv = sys.argv
        # Pipe stdout and stderr to files
        self.__streams = {}
        for name in "stdout", "stderr":
            stream = getattr(sys, name)
            try:
                fileno = stream.fileno()
            except:
                # Already captured by nosetests?
                continue
            outfile = os.path.join(self.__output, name)
            if os.path.exists(outfile):
                os.unlink(outfile)
            new_fd = os.open(outfile,
                             os.O_RDWR | os.O_CREAT | os.O_EXCL,
                             0o600)
            stream.flush()
            saved_fd = os.dup(fileno)
            os.dup2(new_fd, fileno)

            self.__streams[name] = outfile, stream, saved_fd, new_fd

    def tearDown(self):
        # Restore stdout and stderr
        for outfile, stream, saved_fd, new_fd in self.__streams.values():
            stream.flush()
            os.dup2(saved_fd, stream.fileno())
            os.close(saved_fd)
            os.close(new_fd)
        # Restore sys.argv
        sys.argv = self.__saved_argv
        # Delete anything we created in funny places
        for filename in self.__unlink_me:
            if os.path.exists(filename):
                os.unlink(filename)

    def test_factorial(self):
        """Test that the factorial example works."""
        factdir = os.path.join(self.__examples, "factorial")

        source = os.path.join(factdir, "factorial.i8")
        objfile = os.path.join(factdir, "factorial.o")
        testfile = os.path.join(factdir, "test-factorial.py")

        self.__unlink_me.append(objfile)

        # Run the exact commands in the documentation
        sys.argv[1:] = ["-c", source]
        self.assertIs(compiler.main(), None)
        sys.argv[1:] = ["-i", objfile, "-q", "example::factorial(i)i", "12"]
        self.assertIs(runtime.main(), None)
        sys.argv[1:] = ["-i", objfile, testfile]
        self.assertIs(runtime.main(), None)

        # Check nothing was output to stderr
        sys.stderr.flush()
        errors = open(self.__streams["stderr"][0]).read()
        self.assertEqual(errors, "")

        # Check the output of the two I8X invocations
        if hasattr(sys.stdout, "getvalue"):
            # Captured by nosetests
            output = sys.stdout.getvalue()
        else:
            # Captured by us
            sys.stdout.flush()
            output = open(self.__streams["stdout"][0]).read()
        lines = output.split("\n")
        self.assertEqual(lines[0], "479001600")
        self.assertTrue(lines[-4].startswith("Ran 1 test in "))
        self.assertEqual(lines[-2], "OK")



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import constants
from . import logger
from . import RedefinedIdentError
from . import UndefinedIdentError

debug_print = logger.debug_printer_for(__name__)

class Type(object):
    """Base class for all types.
    """
    def __init__(self, ast, name):
        self.ast = ast
        self.name = name

    @property
    def is_builtin(self):
        return self.ast is None

    @property
    def fileline(self):
        assert not self.is_builtin
        return self.ast.fileline

    def lowest_common_ancestor(self, other):
        last = None
        for type1, type2 in zip(self.__lca_path(), other.__lca_path()):
            if type1 is not type2:
                break
            last = type1
        return last

    def __lca_path(self):
        parent = self.basetype
        if parent == self:
            return (self,)
        return self.parent.__lca_path() + (self,)

class BaseType(Type):
    """Base class for types that are not an alias for some other type.
    """
    @property
    def basetype(self):
        return self

    @property
    def sizedtype(self):
        return None

class CoreType(BaseType):
    """The three core types: int, ptr and opaque.
    """
    class_init_complete = False
    is_function = False

    def __init__(self, name, encoding, is_computable):
        assert not CoreType.class_init_complete
        BaseType.__init__(self, None, name)
        self.encoding = encoding
        self.is_computable = is_computable

class FuncType(BaseType):
    """Function types.
    """
    is_computable = False
    is_function = True

    def __init__(self, src, from_function=False):
        if from_function:
            src.parameters.accept(self)
        else:
            src.paramtypes.accept(self)
        src.returntypes.accept(self)
        ptypes = ", ".join((type.name for type in self.paramtypes))
        rtypes = ", ".join((type.name for type in self.returntypes))
        name = "func"
        if rtypes:
            name += " " + rtypes
        name += " (%s)" % ptypes
        Type.__init__(self, src, name)

    @property
    def encoding(self):
        return "%s%s(%s)" % (
            constants.I8_TYPE_FUNC,
            "".join((type.encoding for type in self.returntypes)),
            "".join((type.encoding for type in self.paramtypes)))

    def __eq__(self, other):
        return not (self != other)

    def __ne__(self, other):
        return (not isinstance(other, FuncType)
                or self.paramtypes != other.paramtypes
                or self.returntypes != other.returntypes)

    def visit_parameters(self, parameters):
        self.__list = self.paramtypes = []
        for node in parameters.children:
            node.typename.accept(self)
        del self.__list

    def visit_paramtypes(self, paramtypes):
        self.__list = self.paramtypes = []
        for node in paramtypes.children:
            node.accept(self)
        del self.__list

    def visit_returntypes(self, returntypes):
        self.__list = self.returntypes = []
        for node in returntypes.children:
            node.accept(self)
        del self.__list

    def visit_basictype(self, basictype):
        self.__list.append(basictype.type)

    def visit_functype(self, functype):
        self.__list.append(functype.type)

class AliasType(Type):
    """A type that is an alias for some other type.
    """
    def __init__(self, ast, name, parent):
        assert parent is not None
        Type.__init__(self, ast, name)
        self.parent = parent

    @property
    def basetype(self):
        return self.parent.basetype

    @property
    def sizedtype(self):
        return self.parent.sizedtype

    @property
    def is_computable(self):
        return self.parent.is_computable

    @property
    def is_function(self):
        return self.parent.is_function

    @property
    def encoding(self):
        return self.parent.encoding

class SizedType(AliasType):
    """The sized-integer types used by "deref".
    """
    class_init_complete = False

    def __init__(self, nbits, is_signed):
        assert not SizedType.class_init_complete
        name = "%sint%s_t" % (not is_signed and "u" or "",
                              nbits is None and "ptr" or nbits)
        AliasType.__init__(self, None, name, INTTYPE)
        self.nbits = nbits
        self.is_signed = is_signed

    @property
    def sizedtype(self):
        return self

def __create_builtin_types():
    """Create the builtin types INTTYPE, PTRTYPE, etc.
    """
    def add_builtin_type(type):
        globals()[type.name.upper() + "TYPE"] = type
    for name in ("int", "ptr", "opaque"):
        code = getattr(constants, "I8_TYPE_" + name.upper())
        add_builtin_type(CoreType(name, code, name != "opaque"))
    CoreType.class_init_complete = True
    add_builtin_type(AliasType(None, "bool", INTTYPE))
    for is_signed in range(2):
        add_builtin_type(SizedType(None, is_signed))
        for shift in range(4):
            nbits = 8 << shift
            add_builtin_type(SizedType(nbits, is_signed))
    SizedType.class_init_complete = True
__create_builtin_types()

class TypeAnnotator(object):
    def visit_toplevel(self, toplevel):
        self.types = {}
        for name, type in globals().items():
            if name.endswith("TYPE"):
                self.add_type(type)
        self.in_toplevel = True
        self.newtype = None
        for node in toplevel.children:
            node.accept(self)
        assert self.in_toplevel
        assert self.newtype is None

    def add_type(self, type):
        prev = self.types.get(type.name, None)
        if prev is not None:
            raise RedefinedIdentError(type, "type", type.name, prev)
        self.types[type.name] = type

    def get_type(self, type):
        result = self.types.get(type.name, None)
        if result is None:
            raise UndefinedIdentError(type, "type", type.name)
        return result

    def visit_basictype(self, basictype):
        if self.in_toplevel and self.newtype is not None:
            self.add_basictype(basictype)
        else:
            self.annotate_basictype(basictype)

    def visit_functype(self, functype):
        if self.in_toplevel and self.newtype is not None:
            self.add_functype(functype)
        else:
            self.annotate_functype(functype)

    # Process "define type" statements

    def visit_typedef(self, typedef):
        assert self.newtype is None
        for node in typedef.children:
            node.accept(self)

    def visit_typename(self, typename):
        assert self.newtype is None
        self.newtype = typename

    def add_basictype(self, basictype):
        assert self.newtype is not None
        self.add_type(AliasType(self.newtype,
                                self.newtype.name,
                                self.get_type(basictype)))
        self.newtype = None

    def __annotate_functype(self, functype):
        saved = self.in_toplevel
        try:
            self.in_toplevel = False
            functype.paramtypes.accept(self)
            functype.returntypes.accept(self)
        finally:
            self.in_toplevel = saved

    def add_functype(self, functype):
        assert self.newtype is not None
        self.__annotate_functype(functype)
        self.add_type(AliasType(self.newtype,
                                self.newtype.name,
                                FuncType(functype)))
        self.newtype = None

    # Add "type" fields where necessary in function definitions

    def visit_wordsize(self, wordsize):
        pass

    def visit_function(self, function):
        toplevel_types = self.types
        self.types = self.types.copy()
        self.in_toplevel = False

        for node in function.children:
            node.accept(self)
        function.type = FuncType(function, True)

        if debug_print.is_enabled:
            debug_print("%s\n\n" % function)

        self.types = toplevel_types
        self.in_toplevel = True

    def visit_fullname(self, fullname):
        pass

    def visit_returntypes(self, returntypes):
        for node in returntypes.children:
            node.accept(self)

    def visit_paramtypes(self, paramtypes):
        for node in paramtypes.children:
            node.accept(self)

    def visit_parameters(self, parameters):
        for node in parameters.children:
            node.accept(self)

    def visit_parameter(self, parameter):
        parameter.typename.accept(self)

    def visit_externals(self, externals):
        for node in externals.children:
            node.accept(self)

    def visit_external(self, external):
        assert self.newtype is None
        external.typename.accept(self)

    def annotate_basictype(self, basictype):
        basictype.type = self.get_type(basictype)

    def annotate_functype(self, functype):
        self.__annotate_functype(functype)
        functype.type = FuncType(functype)

    def visit_operations(self, ops):
        for node in ops.children:
            node.accept(self)

    def visit_label(self, label):
        pass

    def visit_operation(self, op):
        pass

    # Visitors for operations that require annotation

    def visit_castop(self, op):
        op.typename.accept(self)

    def visit_derefop(self, op):
        for node in op.children:
            node.accept(self)

    def visit_loadop(self, op):
        for node in op.typed_operands:
            node.accept(self)

    def visit_integer(self, constant):
        constant.type = INTTYPE

    def visit_pointer(self, constant):
        constant.type = PTRTYPE

    def visit_boolean(self, constant):
        constant.type = BOOLTYPE



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import StackError

FUNCTYPE = "func int (int)"
GOODTYPES1 = (FUNCTYPE,)
BADTYPES1 = ("int", "func (int)", "func int (ptr)",
             "func int (int, int)", "func int (int, func int (int))")

SOURCE1 = """\
define test::return_func_test returns %s
    extern func int (int) factorial
    load factorial
"""

ALIASTYPE = "factorial_ft"
GOODTYPES2 = GOODTYPES1 + (ALIASTYPE,)
BADTYPES2 = BADTYPES1

SOURCE2 = ["typedef %s %s" % (FUNCTYPE, ALIASTYPE)]
for index, badtype in zip(range(len(BADTYPES1)), BADTYPES1):
    alias = "wrong_%d" % (index + 1)
    SOURCE2.append("typedef %s %s" % (badtype, alias))
    BADTYPES2 += (alias,)
SOURCE2 = "\n".join(SOURCE2 + [SOURCE1])

class TestReturnFunc(TestCase):
    def __do_test(self, template, goodtypes, badtypes):
        for type in goodtypes + badtypes:
            source = template % type
            if type in goodtypes:
                self.compile(source)
            else:
                self.assertRaises(StackError, self.compile, source)

    def test_raw_return_func(self):
        """Test that returning a function works."""
        self.__do_test(SOURCE1, GOODTYPES1, BADTYPES1)

    def test_typedef_return_func(self):
        """Test that returning a function via a typedef works."""
        self.__do_test(SOURCE2, GOODTYPES2, BADTYPES2)



# -*- coding: utf-8 -*-
# Copyright (C) 2016 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import commands
from . import I8CError
from . import warn
import copy
import struct
import subprocess
import tempfile

class TargetAnnotator(object):
    def __init__(self, commandline):
        self.args = commandline

    def visit_toplevel(self, toplevel):
        # Look for "wordsize" directives in the source.
        self.wordsize = None
        for node in toplevel.wordsize_directives:
            node.accept(self)
        toplevel.wordsize = self.wordsize

        # If no wordsize was specified in the source but we
        # have enough information to run the compiler then
        # we compile an empty file and look at the output.
        if (toplevel.wordsize is None and self.args is not None):
            asmargs = self.__process_sco_args(self.args.asm_args)
            toplevel.wordsize = guess_wordsize(asmargs)
            if toplevel.wordsize is not None:
                if not self.args.with_asm:
                    warn("assuming ‘wordsize %d’" % toplevel.wordsize)

        # If nothing worked we can't continue.
        if toplevel.wordsize is None:
            raise I8CError("unable to determine target wordsize")

    def visit_wordsize(self, wordsize):
        for node in wordsize.children:
            node.accept(self)

    def visit_constant(self, constant):
        assert self.wordsize is None
        self.wordsize = constant.value

    def __process_sco_args(self, src):
        src = copy.copy(src)
        dst = []
        while src:
            arg = src.pop(0)
            if arg == "-o":
                src.pop(0)
            elif arg == "-c" or arg.startswith("-o"):
                continue
            else:
                dst.append(arg)
        return dst

def guess_wordsize(args=None):
    if args is None:
        args = []

    hdrfmt = b"4sB"
    hdrlen = struct.calcsize(hdrfmt)
    try:
        with tempfile.NamedTemporaryFile(suffix=".o") as of:
            with tempfile.NamedTemporaryFile(suffix=".c") as cf:
                command = (commands.I8C_CC
                           + args
                           + ["-c", cf.name, "-o", of.name])
                subprocess.check_call(command)
                with open(of.name, "rb") as fp:
                    header = fp.read(hdrlen)
    except:
        return

    magic, elfclass = struct.unpack(hdrfmt, header)
    if magic == b"\x7fELF":
        return {1: 32, 2: 64}.get(elfclass, None)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

# Simple testcase with only the load constant in the blocks
SIMPLE = """\
typedef int arg_type
typedef int ret_type

define test::optimize_cmp_bra_const_const returns ret_type
    argument arg_type arg1
    argument arg_type arg2

    b%s label

    // comparison returned 0
    load %s
    return

label:
    // comparison returned 1
    load %s
"""

# Harder testcase with an extra operation in each block
HARDER = """\
typedef int arg_type
typedef int ret_type

define test::optimize_cmp_bra_const_const returns ptr, ret_type
    argument ptr arg0
    argument arg_type arg1
    argument arg_type arg2

    b%s label

    // comparison returned 0
    load %s
    swap
    return

label:
    // comparison returned 1
    load %s
    swap
"""

class TestOptimizeCmpBraConstConst(TestCase):
    REVERSE = {"lt": "ge", "le": "gt", "eq": "ne",
               "ne": "eq", "ge": "lt", "gt": "le"}
    OPERATIONS = sorted(REVERSE.keys())

    def test_optimize_cmp_bra_const_const(self):
        """Check that cmp-bra-{const,const} is optimized."""
        # bra is branch-if-NOT-zero
        for source, extra_ops in (SIMPLE, []), (HARDER, ["swap"]):
            for is_reversed, values in ((False, (0, 1)), (True, (1, 0))):
                for op in self.OPERATIONS:
                    tree, output = self.compile(source % ((op,) + values))
                    if is_reversed:
                        op = self.REVERSE[op]
                    self.assertEqual([op] + extra_ops, output.opnames)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

SOURCE = """\
define test::optimize_use_plus_uconst returns int
    argument int x

    load %s
    %s
"""

IDENTITIES = (("add", 0), ("sub", 0),
              ("mul", 1), ("div", 1),
              ("shl", 0), ("shr", 0),
              ("shra", 0), ("or", 0),
              ("xor", 0))

class TestEliminateIdentityMath(TestCase):
    def test_eliminate_identity_math(self):
        """Check that identity math and logic are eliminated."""
        for op, identity in IDENTITIES:
            tree, output = self.compile(SOURCE % (identity, op))
            self.assertEqual(len(output.ops), 0)



# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from i8c.runtime import TestCase

class TestFactorial(TestCase):
    TESTFUNC = "example::factorial(i)i"

    def test_factorial(self):
        """Test example::factorial"""
        for x, expect in ((0, 1), (1, 1), (12, 479001600)):
            result = self.i8ctx.call(self.TESTFUNC, x)
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0], expect)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import integer
from . import parser
from . import visitors
from .types import PTRTYPE

class Operation(visitors.Visitable):
    """Base class for all operations.
    """
    def __init__(self, ast):
        self.ast = ast

    @property
    def fileline(self):
        return self.ast.fileline

    @property
    def source(self):
        """Source file text this operation was created from.
        """
        return " ".join((token.text for token in self.ast.tokens))

    @property
    def is_branch(self):
        return isinstance(self, BranchOp)

    @property
    def is_goto(self):
        return isinstance(self, GotoOp)

    @property
    def is_return(self):
        return isinstance(self, ReturnOp)

    @property
    def is_comparison(self):
        return isinstance(self, CompareOp)

    @property
    def is_load_constant(self):
        return isinstance(self, ConstOp)

    @property
    def is_add(self):
        return isinstance(self, AddOp)

    def __str__(self):
        return '%s("%s")' % (self.__class__.__name__, self.source)

    def is_equivalent_to(self, other):
        if self.is_return and other.is_return:
            return True
        if self.dwarfname != other.dwarfname:
            return False
        assert self.operands == other.operands
        for operand in self.operands:
            if getattr(self, operand) != getattr(other, operand):
                return False
        return True

# All operations either terminate or don't terminate basic blocks.

class TerminalOp(Operation):
    """Base class for all operations that terminate their basic block.
    """
    is_block_terminal = True

class NonTerminalOp(Operation):
    """Base class for all operations that don't terminate basic blocks.
    """
    is_block_terminal = False

    @property
    def dwarfname(self):
        return self.ast.tokens[0].text

# Block-terminating operations.

class BranchOp(TerminalOp):
    BRANCHED_EXIT = 0
    NOBRANCH_EXIT = 1

    dwarfname = "bra"

    @property
    def exit_labels(self):
        yield self.ast.target.name
        yield self.fallthrough

class GotoOp(TerminalOp):
    dwarfname = "skip"

    @property
    def exit_labels(self):
        yield self.ast.target.name

class ReturnOp(TerminalOp):
    exit_labels = ()

# Synthetic block-terminating operations.

class SyntheticGoto(GotoOp):
    def __init__(self, template, target=None):
        assert isinstance(template, Operation)
        GotoOp.__init__(self, parser.SyntheticNode(template.ast,
                                                   "goto"))
        if target is not None:
            self.target = target

    @property
    def exit_labels(self):
        yield self.target

class SyntheticReturn(ReturnOp):
    def __init__(self, template):
        assert isinstance(template, Operation)
        ReturnOp.__init__(self, parser.SyntheticNode(template.ast,
                                                     "return"))
# Non-block-terminating operations.

class NoOperandsOp(NonTerminalOp):
    """Base class for all non-terminal operations without operands.
    """
    operands = ()

# Generic non-block-terminating operations.

class UnaryOp(NoOperandsOp):
    """A math or logic operator that pops one value and pushes one
    back."""
    arity, verb = 1, "operate on"

class BinaryOp(NoOperandsOp):
    """An math or logic operator that pops two values and pushes one
    back."""
    arity, verb = 2, "operate on"

class CompareOp(NoOperandsOp):
    """A comparison operator which pops two values and pushes one
    back."""
    arity, verb = 2, "compare"

    REVERSE = {"lt": "ge", "le": "gt", "eq": "ne",
               "ne": "eq", "ge": "lt", "gt": "le"}

    def __init__(self, ast):
        NonTerminalOp.__init__(self, ast)
        self.reversed = False

    @property
    def dwarfname(self):
        name = self.ast.tokens[0].text[-2:]
        if self.reversed:
            name = self.REVERSE[name]
        return name

    def reverse(self):
        self.reversed = not self.reversed

# Specific non-block-terminating operations.

AbsOp = UnaryOp
AndOp = BinaryOp

class AddOp(BinaryOp):
    dwarfname = "plus"

class PlusUConst(NonTerminalOp):
    dwarfname = "plus_uconst"
    operands = ("value",)

    def __init__(self, template):
        assert template.value != 0
        NonTerminalOp.__init__(self, template.ast)
        self.value = template.value

class CallOp(NoOperandsOp):
    dwarfname = "call"

class CastOp(NonTerminalOp):
    operands = ("slot", "old_type", "new_type")

    @property
    def slot(self):
        return self.ast.slot.value

    @property
    def new_type(self):
        return self.ast.typename.type

class ConstOp(NonTerminalOp):
    operands = ("type", "value")

    @property
    def type(self):
        return self.ast.operand.type

    @property
    def value(self):
        return self.ast.operand.value

class DerefOp(NonTerminalOp):
    arity, verb = 1, "dereference to"
    operands = ("type",)

    @property
    def type(self):
        return self.ast.operand.type

DivOp = BinaryOp

class DropOp(NoOperandsOp):
    pass

class LoadOp(NonTerminalOp):
    def __init__(self, ast):
        NonTerminalOp.__init__(self, ast)
        self.__operand = None

    @property
    def is_resolved(self):
        return self.__operand is not None

    # XXX

    @property
    def is_pick(self):
        assert self.is_resolved
        return isinstance(self.__operand, integer)

    @property
    def is_loadext(self):
        return not self.is_pick

    # Accessors for "pickslot" and "external"

    @property
    def pickslot(self):
        assert self.is_pick
        return self.__operand

    @pickslot.setter
    def pickslot(self, value):
        assert not self.is_resolved
        self.__operand = value
        assert self.is_pick

    @property
    def external(self):
        assert self.is_loadext
        return self.__operand

    @external.setter
    def external(self, value):
        assert not self.is_resolved
        self.__operand = value
        assert self.is_loadext

    # XXX

    @property
    def operands(self):
        if self.is_pick:
            operand = "pickslot"
        else:
            assert self.is_loadext
            operand = "external"
        return (operand,)

    @property
    def dwarfname(self):
        if self.is_pick:
            return {0: "dup", 1: "over"}.get(self.pickslot, "pick")
        assert self.is_loadext
        if self.external.basetype is PTRTYPE:
            return "addr"
        assert self.external.type.is_function
        return "load_external"

    # The thing we are trying to load

    @property
    def name(self):
        return self.ast.name.value

ModOp = BinaryOp
MulOp = BinaryOp

class NameOp(NonTerminalOp):
    operands = ("slot", "newname")

    @property
    def slot(self):
        return self.ast.slot.value

    @property
    def newname(self):
        return self.ast.newname.value

NegOp = UnaryOp
NotOp = UnaryOp
OrOp = BinaryOp

class PickOp(LoadOp):
    def __init__(self, ast, slot):
        LoadOp.__init__(self, ast)
        self.pickslot = slot

ShlOp = BinaryOp
ShrOp = BinaryOp # Welcome to Shropshire
ShraOp = BinaryOp

class SubOp(BinaryOp):
    dwarfname = "minus"

class SwapOp(NoOperandsOp):
    pass

class RotOp(NoOperandsOp):
    pass

class WarnOp(NonTerminalOp):
    dwarfname = "warn"
    operands = ("message",)

    @property
    def value(self):
        return self.ast.operand.value

XorOp = BinaryOp



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys

if sys.version_info < (3,):
    str = unicode
    integer = (int, long,)
    next = lambda i: i.next()
    join_bytes = lambda seq: "".join(c.decode("iso-8859-1")
                                     for c in seq).encode("iso-8859-1")
else:
    str = str
    integer = (int,)
    next = next
    join_bytes = bytes

def fprint(file, text=""):
    fwrite(file, text + "\n")

def fwrite(file, text):
    if sys.version_info < (3,):
        text = text.encode(getattr(file, "encoding", None) or "utf-8",
                           "replace")
    file.write(text)

def strtoint_c(text, exception):
    if text.startswith("-"):
        sign, unsigned = "-", text[1:]
    else:
        sign, unsigned = "", text
    if (len(unsigned) > 1 and unsigned[0] == "0"):
        if unsigned[1] in "oO":
            unsigned = "BadOctal" # Force exception
        elif unsigned[1] not in "xX":
            unsigned = "0o" + unsigned[1:]
    try:
        return int(sign + unsigned, 0)
    except ValueError:
        raise exception("invalid integer literal ‘%s’" % text)

if sys.version_info < (3, 3):
    from imp import load_source as load_module_from_source
else:
    from importlib.machinery import SourceFileLoader

    def load_module_from_source(name, filename):
        return SourceFileLoader(name, filename).load_module()




# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import fprint
from .exceptions import *
from .logger import loggers
import sys

if sys.version_info < (3,):
    str = unicode

def compile(readline, write):
    from .driver import compile
    return compile(readline, write)

def main():
    from .driver import main
    try:
        return main(sys.argv[1:])
    except I8CError as e:
        fprint(sys.stderr, str(e))
        return 1



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import ParserError

SOURCE_0arg = """\
define test::extra_operands_0 returns int
    argument int x
    dup junk
"""

SOURCE_1arg = """\
define test::extra_operands_1 returns int
    argument ptr x
    pick 5, junk
label:
"""

SOURCE_2arg = """\
define test::extra_operands_2 returns int
    argument ptr x
    cast x, int, junk
"""

class TestExtraOperands(TestCase):
    def test_extra_operands(self):
        """Ensure the parser detects extra operands."""
        for source in SOURCE_0arg, SOURCE_1arg, SOURCE_2arg:
            self.assertRaises(ParserError, self.compile, source)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import I8Error
from ..compat import fprint
import sys

def warn(msg, cause=None):
    if cause is None:
        prefix = "i8c"
    else:
        prefix = cause.fileline
    fprint(sys.stderr, "%s: warning: %s" % (prefix, msg))

class I8CError(I8Error):
    """Base class for all compiler errors.
    """
    def __init__(self, msg, prefix="i8c"):
        I8Error.__init__(self, msg, prefix)

class LexerError(I8CError):
    """An error occurred while tokenizing a source file.
    """
    def __init__(self, filename, linenumber, msg):
        I8CError.__init__(self, msg, "%s:%d" % (filename, linenumber))

class LexedError(I8CError):
    """Base class for errors emanating from tokenized source.
    """
    def __init__(self, cause, msg):
        I8CError.__init__(self, msg, cause.fileline)

class ParserError(LexedError):
    """An error occurred while parsing tokens.
    """
    def __init__(self, tokens):
        token = tokens[0]
        LexedError.__init__(self, token, "unexpected ‘%s’" % token.text)

class ParsedError(LexedError):
    """An error occurred while processing the AST.
    """

class IdentError(ParsedError):
    """An error occurred while processing an identifier.
    """

class UndefinedIdentError(IdentError):
    """The specified identifier was not defined.
    """
    def __init__(self, cause, what, name):
        IdentError.__init__(
            self, cause, "%s ‘%s’ is undefined" % (what, name))

class RedefinedIdentError(IdentError):
    """An attempt was made to redefine an identifier.
    """
    def __init__(self, cause, what, name, prev):
        if prev.is_builtin:
            msg = "definition of ‘%s’ shadows builtin %s" % (name, what)
        else:
            msg = ("definition of ‘%s’ shadows previous\n"
                   "%s: error: %s ‘%s’ was previously defined here") % (
                name, prev.fileline, what, name)
        IdentError.__init__(self, cause, msg)

class ReservedIdentError(IdentError):
    """An attempt was made to define an identifier with a reserved name.
    """
    def __init__(self, cause, what, name):
        IdentError.__init__(
            self, cause, "%s ‘%s’ is reserved" % (what, name))

class BlockCreatorError(ParsedError):
    """An error occurred creating basic blocks from the AST.
    """

class StackError(ParsedError):
    """The stack is not correct for the requested operation.
    """
    def __init__(self, cause, stack, msg):
        msg = "%s: %s" % (cause.source, msg)
        if stack is not None:
            msg += ":\n" + str(stack)
        ParsedError.__init__(self, cause, msg)

class StackTypeError(StackError):
    """The stack contains incorrect types for the requested operation.
    """
    def __init__(self, cause, stack):
        # Sorry translators...
        types = ["‘%s’" % stack[index].type.name
                 for index in range(cause.arity - 1, -1, -1)]
        if len(types) > 1:
            types[:-1] = [", ".join(types[:-1])]
        types = " and ".join(types)
        msg = "can't %s %s" % (cause.verb, types)
        StackError.__init__(self, cause, stack, msg)

class StackMergeError(StackError):
    def __init__(self, ops, stacks, slot=None):
        from_op, to_op = ops
        prev_stack, new_stack = stacks
        msg = "\narriving from %s with stack:\n%s\n" % (
            from_op.fileline, new_stack)
        msg += "can't merge with previously walked stack:\n%s\n" % prev_stack
        if slot is not None:
            msg += "because of incompatible type at stack[%d]." % slot
        else:
            msg += "because depths differ"
        StackError.__init__(self, to_op, None, msg)

class CastError(StackError):
    """A cast operation is somehow wrong.
    """

class UnnecessaryCastError(CastError):
    """An unnecessary cast was encountered.
    """
    def __init__(self, cause, stack):
        CastError.__init__(self, cause, stack, "unnecessary ‘cast’")

class InvalidCastError(CastError):
    """An invalid cast was encountered.
    """
    def __init__(self, cause, stack, old_type, new_type):
        msg = "can't cast from ‘%s’ to ‘%s’" % (old_type.name,
                                                new_type.name)
        CastError.__init__(self, cause, stack, msg)




# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import parser
from i8c.compiler.types import INTTYPE, PTRTYPE, BOOLTYPE

INPUT_TEST = """\
define test::input_test
    load 23
    load 023
    load 0x23
    load -17
    load -017
    load -0x17
    load NULL
    load TRUE
    load FALSE
"""

class TestLoadConstantInput(TestCase):
    def test_input(self):
        """Check that the input is parsed correctly.
        """
        tree, output = self.compile(INPUT_TEST)
        constants = []
        node = tree.one_child(parser.Function)
        node = node.one_child(parser.Operations)
        for op in node.children:
            self.assertIsInstance(op, parser.LoadOp)
            for node in op.children:
                self.assertIsInstance(node, parser.Constant)
                constants.append([node.type, node.value])
        self.assertEqual([[INTTYPE, 23],
                          [INTTYPE, 0o23],
                          [INTTYPE, 0x23],
                          [INTTYPE, -17],
                          [INTTYPE, -0o17],
                          [INTTYPE, -0x17],
                          [PTRTYPE, 0],
                          [BOOLTYPE, 1],
                          [BOOLTYPE, 0]], constants)

OUTPUT_TEST = (
    (0, "lit0"),
    (31, "lit31"),
    (32, "const1u"),
    (255, "const1u"),
    (256, "const2u"),
    (65535, "const2u"),
    (65536, "constu"),
    (2097151, "constu"),
    (2097152, "const4u"),
    (4294967295, "const4u"),
    (4294967296, "constu"),
    (562949953421311, "constu"),
    (562949953421312, "const8u"),
    (18446744073709551615, "const8u"),
    (18446744073709551616, "constu"),
    (-1, "const1s"),
    (-128, "const1s"),
    (-129, "const2s"),
    (-32768, "const2s"),
    (-32769, "consts"),
    (-1048576, "consts"),
    (-1048577, "const4s"),
    (-2147483648, "const4s"),
    (-2147483649, "consts"),
    (-281474976710656, "consts"),
    (-281474976710657, "const8s"),
    (-9223372036854775808, "const8s"),
    (-9223372036854775809, "consts"))

class TestLoadConstantOutput(TestCase):
    def test_output(self):
        """Check that the correct bytecodes are emitted.
        """
        for value, opname in OUTPUT_TEST:
            tree, output = self.compile(
                "define test::input_test\nload %d" % value)

            ops = output.ops
            self.assertEqual(len(ops), 1)
            op = ops[0]
            self.assertEqual(op.name, opname)
            if value < 0 or value > 31:
                self.assertEqual(op.operand, value)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

class I8Error(Exception):
    """Base class for all errors.
    """
    def __init__(self, msg, prefix):
        Exception.__init__(self, prefix + ": error: " + msg)

def version():
    try:
        import pkg_resources
        return pkg_resources.get_distribution("i8c").version
    except: # pragma: no cover
        # This block is excluded from coverage because while
        # we could test it (by hiding the egg-info somehow?)
        # it seems like a lot of effort for very little gain.
        return "UNKNOWN"



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import StackError

SOURCE = """\
define test::comparison_test
    argument %s arg_1
    argument %s arg_2

    %s
label:
"""

class TestComparisons(TestCase):
    OPERATIONS = "lt", "le", "eq", "ne", "ge", "gt"
    TYPES = "int", "bool", "ptr", "opaque", "func ()"

    def test_comparisons(self):
        """Basic checks for compare and compare+branch bytecodes."""
        for prefix, suffix in (("", ""), ("b", " label")):
            for op in self.OPERATIONS:
                for type1 in self.TYPES:
                    for type2 in self.TYPES:
                        expect_ops = [op]
                        if prefix == "b":
                            expect_ops.append("bra")
                        self.__run_test(SOURCE % (type1, type2,
                                                  prefix + op + suffix),
                                        self.__expect_success(type1, type2),
                                        expect_ops)

    def __expect_success(self, type1, type2):
        if type1 == "ptr" and type2 == "ptr":
            return True
        if type1 in ("int", "bool") and type2 in ("int", "bool"):
            return True
        return False

    def __run_test(self, source, expect_success, expect_ops):
        if expect_success:
            tree, output = self.compile(source)
            self.assertEqual(expect_ops, output.opnames)
        else:
            self.assertRaises(StackError, self.compile, source)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import fprint
from .exceptions import *
from .context import Context
from .testcase import TestCase
import sys

if sys.version_info < (3,):
    str = unicode

def main():
    from .driver import main
    try:
        return main(sys.argv[1:])
    except I8XError as e:
        fprint(sys.stderr, str(e))
        return 1



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import integer
from . import InvalidCastError
from . import logger
from . import names
from . import ParsedError
from . import RedefinedIdentError
from . import StackError
from . import StackMergeError
from . import StackTypeError
from . import UndefinedIdentError
from . import UnnecessaryCastError
from .types import Type, INTTYPE, PTRTYPE, BOOLTYPE
import copy

debug_print = logger.debug_printer_for(__name__)

class CastError(Exception):
    def __init__(self, *args):
        self.args = args

    def decorate_and_raise(self, *args):
        raise self.EXCEPTION(*(args + self.args))

class _UnnecessaryCastError(CastError):
    EXCEPTION = UnnecessaryCastError

class _InvalidCastError(CastError):
    EXCEPTION = InvalidCastError

class Stack(object):
    def __init__(self, externals):
        self.externals = externals
        self.slots = []
        self.is_mutable = True
        self.max_depth = 0

    @property
    def default_provider(self):
        return self.externals.default_provider

    @property
    def depth(self):
        return len(self.slots)

    def push(self, elem):
        assert self.is_mutable
        assert isinstance(elem, Element)
        self.slots.insert(0, elem)
        self.max_depth = max(self.max_depth, self.depth)

    def __push_back(self, item):
        assert self.is_mutable
        self.slots.append(item)
        self.max_depth = max(self.max_depth, self.depth)

    def pop(self):
        assert self.is_mutable
        self.underflow_check(0)
        return self.slots.pop(0)

    def name_slot(self, name_or_index, newname):
        assert self.is_mutable
        index = self.__get_name_cast_index(name_or_index)

        slots = self.__slots_matching(newname)
        if slots:
            if index in slots:
                return # slot already has this name
            self.__raise_redef_error(newname, self[slots[0]])
        external = self.externals.lookup(newname)
        if external is not None:
            self.__raise_redef_error(newname, external)

        elem = copy.copy(self.slots[index])
        elem.names = copy.copy(elem.names)
        elem.names.append(newname)
        self.slots[index] = elem

    def __raise_redef_error(self, newname, prev):
        assert newname.is_shortname
        for prev in prev.names:
            if prev.name == newname.name:
                break
        else:
            raise AssertionError
        raise RedefinedIdentError(newname, "name", newname.name, prev)

    def cast_slot(self, name_or_index, new_type):
        assert self.is_mutable
        assert isinstance(new_type, Type)
        index = self.__get_name_cast_index(name_or_index)
        elem = copy.copy(self.slots[index])
        old_type = elem.type

        if new_type.basetype == old_type.basetype:
            raise _UnnecessaryCastError()
        if not old_type.is_computable or not new_type.is_computable:
            raise _InvalidCastError(old_type, new_type)

        elem.type = new_type
        self.slots[index] = elem
        return old_type

    def __get_name_cast_index(self, name_or_index):
        if isinstance(name_or_index, integer):
            self.underflow_check(name_or_index)
            return name_or_index
        else:
            assert isinstance(name_or_index, names.Name)
            slots = self.__slots_matching(name_or_index)
            if len(slots) != 1:
                raise UndefinedIdentError(
                    name_or_index, "name", str(name_or_index))
            return slots[0]

    def get_by_name(self, name):
        slots = self.__slots_matching(name)
        if len(slots) == 1:
            return slots[0]
        external = self.externals.lookup(name)
        if external is None:
            raise UndefinedIdentError(name, "name", str(name))
        return external

    def __slots_matching(self, name):
        if name.is_shortname:
            search = [name, name.with_provider(self.default_provider)]
        else:
            search = [name]
        results = []
        for elem, index in zip(self.slots, range(self.depth)):
            # Does this slot match the names we're looking for?
            if not self.__names_match(search, elem.names):
                continue
            # Is this slot just a copy of a previous result?
            if elem in (self.slots[result] for result in results):
                continue
            # This is a new match
            results.append(index)
        return results

    def __names_match(self, list1, list2):
        for name in list1:
            if name in list2:
                return True
        return False

    def make_immutable(self):
        self.is_mutable = False

    def mutable_copy(self):
        assert not self.is_mutable
        # Don't use deepcopy.  We want to copy the slots list
        # but copying elements is wasteful (we don't need it)
        # and copying element types is problematic as it breaks
        # "is" comparison.
        result = copy.copy(self)
        result.slots = copy.copy(self.slots)
        result.is_mutable = True
        return result

    def underflow_check(self, depth):
        assert depth >= 0
        if self.depth <= depth:
            raise StackError(self.current_op, None, "stack underflow")

    def __getitem__(self, index):
        self.underflow_check(index)
        return self.slots[index]

    def __str__(self):
        if not self.slots:
            return "   <empty stack>"
        return "\n".join(("%4d: %s" % (slot, self[slot])
                          for slot in range(self.depth)))

    def merge_into(self, previous, ops):
        # There is a direction to this.  We (self) are a new stack
        # joining anoterh, previously processed stack.  If no merge
        # is necessary we indicate this by return the previous stack
        # as our result (i.e. the result of the merge is no change,
        # so the block does not need re-walking.

        if previous is self:
            return previous

        if previous.depth != self.depth:
            raise StackMergeError(ops, (previous, self))

        merged = Stack(self.externals)
        for slot in range(self.depth):
            selem = self[slot]
            pelem = previous[slot]

            if pelem is selem:
                merged.__push_back(pelem)
                continue

            # Merge the types.
            merged_type = pelem.type.lowest_common_ancestor(selem.type)
            if merged_type is None:
                raise StackMergeError(ops, (previous, self), slot)

            # Merge the values.
            if pelem.value == selem.value:
                merged_value = pelem.value
            else:
                merged_value = None

            # Merge the names.
            merged_names = []
            for name in pelem.names:
                if name in selem.names:
                    merged_names.append(name)

            if (merged_type == pelem.type
                and merged_value == pelem.value
                and merged_names == pelem.names):
                merged_elem = pelem
            else:
                merged_elem = Element(merged_type, None, merged_value)
                merged_elem.names = merged_names

            merged.__push_back(merged_elem)

        if merged.slots == previous.slots:
            return previous

        merged.is_mutable = previous.is_mutable
        merged.max_depth = max(previous.max_depth, self.max_depth)
        return merged

class Element(object):
    def __init__(self, thetype, name=None, value=None):
        assert thetype is not None
        assert isinstance(thetype, Type)
        self.type = thetype
        self.names = []
        if name is not None:
            assert isinstance(name, names.Name)
            self.names.append(name)
        if value is not None:
            assert isinstance(value, integer)
        self.value = value

    @property
    def basetype(self):
        return self.type.basetype

    def __str__(self):
        result = ""
        if not self.names:
            result += "anonymous "
        result += self.type.name
        if self.names:
            result += " "
            if len(self.names) > 1:
                result += "["
            result += ", ".join(map(str, self.names))
            if len(self.names) > 1:
                result += "]"
        if self.value is not None:
            result += " = %d" % self.value
        return result

class StackWalker(object):
    def visit_toplevel(self, toplevel):
        for node in toplevel.functions:
            node.accept(self)

    def visit_function(self, function):
        # Build the entry stack
        self.entry_stack = Stack(function.externals)
        function.parameters.accept(self)
        self.entry_stack.make_immutable()
        # Build the return types
        self.returntypes = []
        function.returntypes.accept(self)
        # Walk the operations
        self.max_stack = 0
        self.__enter_block(function.entry_block, None, self.entry_stack)
        function.max_stack = self.max_stack

    # Build the entry stack

    def visit_parameters(self, parameters):
        for node in parameters.children:
            node.accept(self)

    def visit_parameter(self, param):
        type = param.typename.type
        name = param.name.value
        self.entry_stack.push(Element(type, name))

    # Build the return types

    def visit_returntypes(self, returntypes):
        for node in returntypes.children:
            node.accept(self)

    def visit_basictype(self, basictype):
        self.returntypes.append(basictype.type)

    def visit_functype(self, functype):
        self.returntypes.append(functype.type)

    # Manage transitions between blocks

    def __enter_block(self, block, from_block, new_entry_stack):
        self.__current_block = block
        old_entry_stack = getattr(block, "entry_stack", None)
        if old_entry_stack is not None:
            new_entry_stack = new_entry_stack.merge_into(
                old_entry_stack, (from_block.last_op,
                                  block.first_op))
            if new_entry_stack is old_entry_stack:
                return # We've walked this block with this stack
        block.entry_stack = new_entry_stack
        block.accept(self)

    def __leave_block(self):
        from_block = self.__current_block
        stack = self.stack
        del self.__current_block, self.stack
        stack.make_immutable()
        self.max_stack = max(self.max_stack, stack.max_depth)
        for block in from_block.exits:
            self.__enter_block(block, from_block, stack)

    # Walk the operations

    def visit_basicblock(self, block):
        self.stack = block.entry_stack.mutable_copy()
        debug_print("%s:\n\n%s\n" % (block.name, self.stack))
        for op in block.ops:
            debug_print("\n %s\n\n" % op)
            self.stack.current_op = op
            op.accept(self)
            if hasattr(self, "stack"):
                debug_print("%s\n" % self.stack)

    def visit_addop(self, op):
        # Check the types before mutating the stack
        # so any error messages show the whole setup
        fulltypes = [self.stack[index].type for index in (0, 1)]
        basetypes = [type.basetype for type in fulltypes]
        # At least one of the operands must be INTTYPE
        if INTTYPE not in basetypes:
            raise StackTypeError(op, self.stack)
        if PTRTYPE in basetypes:
            rtype = PTRTYPE
        else:
            rtype = fulltypes[0].lowest_common_ancestor(fulltypes[1])
            if rtype is None:
                # One of the types was not INTTYPE
                raise StackTypeError(op, self.stack)
        # Now pop everything and push the result
        self.stack.pop()
        self.stack.pop()
        self.stack.push(Element(rtype))

    def visit_branchop(self, op):
        assert self.stack[0].type is BOOLTYPE
        self.stack.pop()
        self.__leave_block()

    # and, div, mod, mul, or, shl, shr, shra, xor
    def visit_binaryop(self, op):
        # Check the types before mutating the stack
        # so any error messages show the whole setup
        fulltypes = [self.stack[index].type for index in (0, 1)]
        basetypes = [type.basetype for type in fulltypes]
        if basetypes != [INTTYPE, INTTYPE]:
            raise StackTypeError(op, self.stack)
        rtype = fulltypes[0].lowest_common_ancestor(fulltypes[1])
        assert rtype is not None
        self.stack.pop()
        self.stack.pop()
        self.stack.push(Element(rtype))

    def visit_callop(self, op):
        # Check the types before mutating the stack
        # so any error messages show the whole setup
        ftype = self.stack[0].basetype
        if not ftype.is_function:
            raise StackError(op, self.stack, "stack[0] not a function:")
        num_params = len(ftype.paramtypes)
        self.stack.underflow_check(num_params)
        for pindex in range(num_params):
            sindex = num_params - pindex
            ptype = ftype.paramtypes[pindex]
            stype = self.stack[sindex].type
            if stype.basetype != ptype.basetype:
                raise StackError(op, self.stack,
                                 "wrong type in stack[%d]" % sindex)
        # Now pop everything and push the result
        for i in range(1 + num_params):
            self.stack.pop()
        num_returns = len(ftype.returntypes)
        for sindex in range(num_returns):
            rindex = num_returns - sindex - 1
            self.stack.push(Element(ftype.returntypes[rindex]))

    def visit_castop(self, op):
        try:
            op.old_type = self.stack.cast_slot(op.slot, op.new_type)
        except CastError as e:
            e.decorate_and_raise(op, self.stack)

    def visit_compareop(self, op):
        # Check the types before mutating the stack
        # so any error messages show the whole setup
        typea = self.stack[0].type
        typeb = self.stack[1].type
        for type in typea, typeb:
            if not type.is_computable:
                raise StackTypeError(op, self.stack)
        if typea.basetype is not typeb.basetype:
            raise StackTypeError(op, self.stack)
        # Now pop everything and push the result
        self.stack.pop()
        self.stack.pop()
        self.stack.push(Element(BOOLTYPE))

    def visit_constop(self, op):
        self.stack.push(Element(op.type, value=op.value))

    def visit_derefop(self, op):
        rtype = op.type
        if rtype.basetype is not PTRTYPE:
            sizedtype = rtype.sizedtype
            if sizedtype is None:
                raise ParsedError(
                    op, "can't ‘deref’ unsized type ‘%s’" % rtype.name)
        # Check the types before mutating the stack
        # so any error messages show the whole setup
        type = self.stack[0].type
        if not type.basetype is PTRTYPE:
            raise StackTypeError(op, self.stack)
        self.stack.pop()
        self.stack.push(Element(rtype))

    def visit_dropop(self, op):
        self.stack.pop()

    def visit_gotoop(self, op):
        self.__leave_block()

    def visit_loadop(self, op):
        item = self.stack.get_by_name(op.name)
        if isinstance(item, integer):
            if not op.is_resolved:
                op.pickslot = item
            else:
                assert item == op.pickslot
            item = self.stack[item]
        else:
            if not op.is_resolved:
                op.external = item
            else:
                assert item == op.external
        self.stack.push(item)

    def visit_nameop(self, op):
        self.stack.name_slot(op.slot, op.newname)

    def visit_pickop(self, op):
        self.stack.push(self.stack[op.pickslot])

    def visit_subop(self, op):
        # Check the types before mutating the stack
        # so any error messages show the whole setup
        fulltypes = [self.stack[index].type for index in (0, 1)]
        basetypes = [type.basetype for type in fulltypes]
        if basetypes == [INTTYPE, PTRTYPE]:
            rtype = PTRTYPE
        elif basetypes == [INTTYPE, INTTYPE]:
            rtype = fulltypes[0].lowest_common_ancestor(fulltypes[1])
            assert rtype is not None
        else:
            raise StackTypeError(op, self.stack)
        self.stack.pop()
        self.stack.pop()
        self.stack.push(Element(rtype))

    def visit_returnop(self, op):
        num_returns = len(self.returntypes)
        for index in range(num_returns):
            rtype = self.returntypes[index].basetype
            stype = self.stack[index].basetype
            if rtype != stype:
                raise StackError(op, self.stack,
                                 "wrong type in stack[%d]" % index)
        self.__leave_block()

    def visit_rotop(self, op):
        a = self.stack.pop()
        b = self.stack.pop()
        c = self.stack.pop()
        self.stack.push(a)
        self.stack.push(c)
        self.stack.push(b)

    # abs, neg, not
    def visit_unaryop(self, op):
        # Check the types before mutating the stack
        # so any error messages show the whole setup
        if self.stack[0].basetype is not INTTYPE:
            raise StackTypeError(op, self.stack)
        a = self.stack.pop()
        self.stack.push(Element(a.type))

    def visit_swapop(self, op):
        a = self.stack.pop()
        b = self.stack.pop()
        self.stack.push(a)
        self.stack.push(b)

    def visit_warnop(self, op):
        pass



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

SOURCE = """\
define test::last_op_is_branch
    argument bool x
    argument bool y

    goto label2
label1:
    return

label2:
    bne label1
"""

class TestFuncWithLastOpBra(TestCase):
    def test_last_op_is_branch(self):
        """Check that functions whose last op is a branch work.

        This is testing the code that adds the synthetic return.
        As a side-effect it also exercises the code that stops
        us generating unnecessary gotos.
        """
        tree, output = self.compile(SOURCE)
        self.assertEqual(["ne", "bra"], output.opnames)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import ParserError

FUNCSRC = """\
define test::test_toplevel_extern_func returns int
    argument ptr p
    call a_function
"""

FUNCSRC1 = """\
extern func int, int (ptr) %s
""" + FUNCSRC

FUNCSRC2 = """\
typedef func int, int (ptr) functype
extern functype %s
""" + FUNCSRC

class TestTopLevelExtFunc(TestCase):
    def test_basic(self):
        """Check that basic toplevel external functions work."""
        self.__test_good(FUNCSRC1)

    def test_typedef(self):
        """Check that typedef toplevel external functions work."""
        self.__test_good(FUNCSRC2)

    def __test_good(self, source):
        tree, output = self.compile(source % "test::a_function")
        self.assertEqual(["load_external", "call"], output.opnames)

    def test_basic_no_prov(self):
        """Check basic toplevel external functions without providers fail."""
        self.__test_no_prov(FUNCSRC1)

    def test_typedef_no_prov(self):
        """Check typedef toplevel external functions without providers fail."""
        self.__test_no_prov(FUNCSRC2)

    def __test_no_prov(self, source):
        self.assertRaises(ParserError, self.compile, source % "a_function")


SYMSRC = """\
define test::test_toplevel_extern_sym returns ptr
    load a_symbol
"""

SYMSRC1 = """\
extern ptr %s
""" + SYMSRC

SYMSRC2 = """\
typedef ptr symtype
extern symtype %s
""" + SYMSRC

class TestTopLevelExtSym(TestCase):
    def test_basic(self):
        """Check that basic toplevel symbols work."""
        self.__test_good(SYMSRC1)

    def test_typedef(self):
        """Check that typedef toplevel symbols work."""
        self.__test_good(SYMSRC2)

    def __test_good(self, source):
        tree, output = self.compile(source % "a_symbol")
        self.assertEqual(["addr"], output.opnames)

    def test_basic_with_prov(self):
        """Check basic toplevel symbols with providers fail."""
        self.__test_with_prov(SYMSRC1)

    def test_typedef_with_prov(self):
        """Check typedef toplevel symbols with providers fail."""
        self.__test_with_prov(SYMSRC2)

    def __test_with_prov(self, source):
        self.assertRaises(ParserError, self.compile,
                          source % "test::a_symbol")





# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

SOURCE = """\
define test::combine_blocks returns int
    argument int arg1
    argument int arg2

    load arg1
    bge 5, split_1a
    goto split_1b

split_1a:
    load arg2
    blt 4, split_2a
    goto split_2b

split_1b:
    load arg2
    bgt 2, split_2c
    goto split_2d

split_2a:
    load 23
    return

split_2b:
    load 16
    return

split_2c:
    load 16
    return

split_2d:
    load arg1
    add arg2
    bge 1, split_3
    goto split_1a

split_3:
    load 23
    return
"""

class TestCombineBlocks(TestCase):
    def test_combine_blocks(self):
        """Basic block-combiner test."""
        tree, output = self.compile(SOURCE)
        ops = output.opnames
        # Robust test: check the stream contains only one lit16
        # and only one lit23.
        self.assertEqual(len([op for op in ops if op == "lit16"]), 1)
        self.assertEqual(len([op for op in ops if op == "lit23"]), 1)
        # Fragile test: check that the stream is as we expect.
        # This may need fixing up to cope with future compiler changes.
        self.assertEqual(["over", "lit5", "lt", "bra", "dup", "lit4",
                          "ge", "bra", "lit23", "skip", "lit16", "skip",
                          "dup", "lit2", "le", "bra", "skip", "over",
                          "over", "plus", "lit1", "lt", "bra", "skip"],
                         output.opnames)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import I8CError
from i8c.compiler.driver import main
import os
import subprocess
import sys

SOURCE = """\
define test::func
    return
"""

class TestCompilerDriver(TestCase):
    """Test i8c.compiler.driver.main.

    This testcase should be the bare minumum to exercise the function
    i8c.compiler.driver.main and its helper functions.  Command line
    processing tests should be in test_commandline_processor.py, and
    tests exercising the compiler generally (i8c.compiler.compile)
    should be in their own files.
    """

    def __workdir(self):
        test_id = self.id().split(".")
        # Remove the common prefix and the name of the class
        assert test_id[0] == "tests"
        test_id.pop(0)
        test_id.pop(-2)
        return os.path.join(self.topdir, "tests", "output", *test_id)

    def setUp(self):
        # Set up a working directory
        self.workdir = self.__workdir()
        if os.path.exists(self.workdir):
            os.chmod(self.workdir, 0o700)
            subprocess.call(("rm", "-rf", self.workdir))
        os.makedirs(self.workdir)
        self.filebase = os.path.join(self.workdir, "test")
        self.infile = self.filebase + ".i8"
        with open(self.infile, "w") as fp:
            fp.write(SOURCE)
        # Pipe stderr to a file
        tmpfile = os.path.join(self.workdir, "stderr")
        self.stderr_fd = os.open(tmpfile,
                                 os.O_RDWR | os.O_CREAT | os.O_EXCL,
                                 0o600)
        sys.stderr.flush()
        self.saved_stderr_fd = os.dup(2)
        os.dup2(self.stderr_fd, 2)

    def tearDown(self):
        # Restore stderr
        sys.stderr.flush()
        os.dup2(self.saved_stderr_fd, 2)
        os.close(self.saved_stderr_fd)
        os.close(self.stderr_fd)

    # Test all specifiable permutations of (with_cpp,with_i8c,with_asm)

    def __run_permtest(self, args, outext):
        self.outfile = self.filebase + outext
        if "-E" in args:
            args.extend(("-o", self.outfile))
        args.append(self.infile)
        self.assertFalse(os.path.exists(self.outfile))
        status = main(args)
        self.assertIs(status, None)
        self.assertTrue(os.path.isfile(self.outfile))
        junk = os.path.join(self.workdir, "-.o")
        self.assertFalse(os.path.exists(junk))

    def test_do_nothing(self):
        """Check that -E -fpreprocessed is rejected."""
        self.assertRaises(I8CError, main, ["-E", "-fpreprocessed"])

    def test_pp_to_asm(self):
        """Check that preprocessed source to assembly works."""
        self.__run_permtest(["-S", "-fpreprocessed"], ".S")

    def test_pp_to_wrap_asm(self):
        """Check that preprocessed source to wrapped assembly works."""
        self.__run_permtest(["-S", "-fpreprocessed", "--wrap-asm"], ".c")

    def test_pp_to_obj(self):
        """Check that preprocessed source to object code works."""
        self.__run_permtest(["-fpreprocessed", "-c"], ".o")

    def test_i8_to_pp(self):
        """Check that i8 source to preprocessed source works."""
        self.__run_permtest(["-E"], ".i8p")

    def test_i8_to_asm(self):
        """Check that i8 source to assembly works."""
        self.__run_permtest(["-S"], ".S")

    def test_i8_to_wrap_asm(self):
        """Check that i8 source to wrapped assembly works."""
        self.__run_permtest(["-S", "--wrap-asm"], ".c")

    def test_i8_to_obj(self):
        """Check that i8 source to object code works."""
        self.__run_permtest(["-c"], ".o")

    # Test that GCC errors are handled correctly

    def __run_failtest(self):
        status = main(["-c", self.infile])
        self.assertIsNot(status, None)
        size = os.lseek(self.stderr_fd, 0, 1)
        os.lseek(self.stderr_fd, 0, 0)
        output = os.read(self.stderr_fd, size).decode("utf-8")
        self.assertGreaterEqual(output.find("error:"), 0)

    def test_cpp_failure(self):
        """Check that preprocessor errors are handled correctly."""
        os.unlink(self.infile)
        self.__run_failtest()

    def test_asm_failure(self):
        """Check that assembler errors are handled correctly."""
        os.chmod(self.workdir, 0o500)
        try:
            self.__run_failtest()
        finally:
            os.chmod(self.workdir, 0o700)

    # Test that multiple input files with no output file is caught

    def test_multi_input_no_output(self):
        """Check that unguessable output filenames are handled."""
        infile2 = os.path.join(self.workdir, "test2.i8")
        open(infile2, "w")
        self.assertRaises(I8CError,
                          self.__run_permtest, ["-c", infile2], ".o")



# -*- coding: utf-8 -*-
# Copyright (C) 2016 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

def encode(wordsize):
    assert wordsize in range(256)
    msb = ord("i") ^ wordsize
    lsb = ord("8") ^ wordsize
    return (msb << 8) | lsb



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase

SOURCE1 = """\
define test::optimize_cmp_bra_const_const returns ptr
    argument ptr x

    dup
    load NULL
    beq return_the_null
    deref ptr
    return
return_the_null:
"""

SOURCE2 = """\
define test::optimize_cmp_bra_const_const returns ptr
    argument ptr x

    dup
    load NULL
    bne dereference
    return
dereference:
    deref ptr
"""

class TestOptimizeLit0CmpBeforeBra(TestCase):
    def test_optimize_lit0_cmp_before_bra(self):
        """Check that lit0,cmp before bra is eliminated."""
        for source in SOURCE1, SOURCE2:
            tree, output = self.compile(source)
            self.assertEqual(["dup", "bra", "skip", "deref"], output.opnames)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Execution Environment.
#
# The Infinity Note Execution Environment is free software; you can
# redistribute it and/or modify it under the terms of the GNU Lesser
# General Public License as published by the Free Software Foundation;
# either version 2.1 of the License, or (at your option) any later
# version.
#
# The Infinity Note Execution Environment is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with the Infinity Note Execution Environment; if not,
# see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..compat import fprint, str
from . import *
from . import provider
from . import functions
from . import stack
import sys

class Context(object):
    def __init__(self):
        self.functions = {}
        self.env = None
        self.wordsize = None
        self.byteorder = None
        self.tracelevel = 0
        self.__last_traced = None

    # Methods to XXX

    def register_function(self, function):
        funclist = self.functions.get(function.signature, [])
        if not funclist:
            self.functions[function.signature] = funclist
        funclist.append(function)

    def get_function(self, sig_or_ref):
        if isinstance(sig_or_ref, functions.UnresolvedFunction):
            reference = sig_or_ref
            signature = reference.signature
        else:
            if isinstance(sig_or_ref, bytes):
                sig_or_ref = sig_or_ref.decode("utf-8")
            signature = sig_or_ref
            reference = None
        assert isinstance(signature, str)
        # First check for implementations in the testcase.
        # We do this first to allow testcases to override
        # functions which may actually exist.
        if reference is not None:
            impl = "%s_%s_impl" % (reference.provider, reference.name)
            impl = getattr(self.env, impl, None)
            if impl is not None:
                return functions.BuiltinFunction(reference, impl)
        # Now check the registered functions
        funclist = self.functions.get(signature, None)
        if funclist is not None:
            if len(funclist) == 1:
                return funclist[0]
            elif len(funclist) > 1:
                raise AmbiguousFunctionError(sig_or_ref)
        # No registered function with this name
        raise UndefinedFunctionError(sig_or_ref)

    # Methods to XXX

    def import_notes(self, filename):
        with provider.open(filename) as np:
            for note in np.infinity_notes:
                self.import_note(note)

    def import_note(self, note):
        if self.wordsize is None:
            self.wordsize = note.wordsize
        else:
            assert note.wordsize == self.wordsize
        if self.byteorder is None:
            self.byteorder = note.byteorder
        else:
            assert note.byteorder == self.byteorder
        self.register_function(functions.BytecodeFunction(note))

    def new_stack(self):
        return stack.Stack(self.wordsize)

    def call(self, signature, *args):
        function = self.get_function(signature)
        stack = self.new_stack()
        stack.push_multi(reversed(function.ptypes), reversed(args))
        function.execute(self, stack)
        return stack.pop_multi(function.rtypes)

    def __trace(self, location, stack, encoded, decoded):
        function, pc = location
        if self.tracelevel > 0:
            if function != self.__last_traced:
                fprint(sys.stdout, "\n%s:" % function)
                self.__last_traced = function
            if self.tracelevel > 1:
                stack.trace(self.tracelevel)
            fprint(sys.stdout, "  %04x: %-12s %s" % (pc, encoded, decoded))

    def trace_operation(self, *args):
        self.__trace(*args)

    def trace_call(self, function, stack):
        if not isinstance(function, functions.BytecodeFunction):
            if self.tracelevel > 0:
                fprint(sys.stdout, "\n%s:" % function)
                fprint(sys.stdout, "  NON-BYTECODE FUNCTION")
        self.__last_traced = None

    def trace_return(self, location, stack):
        self.__trace(location, stack, "", "RETURN")
        self.__last_traced = None



# -*- coding: utf-8 -*-
# Copyright (C) 2015-16 Red Hat, Inc.
# This file is part of the Infinity Note Compiler.
#
# The Infinity Note Compiler is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Infinity Note Compiler is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the Infinity Note Compiler.  If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tests import TestCase
from i8c.compiler import UndefinedIdentError, StackMergeError

# A basic iterative factorial function.
SOURCE_BASIC = """\
define test::factorial returns int
    argument int x

    load 1
    swap
    goto check

loop:
    dup
    rot
    mul
    swap
    load 1
    sub

check:
    dup
    load 1
    bgt loop
    drop
"""

# A broken version of SOURCE_BASIC which falls through to "check"
# with an extra element on the stack that should be caught by the
# merge.
SOURCE_BAD_DEPTH = """\
define test::factorial returns int
    argument int x

    load 1
    swap
    goto check

loop:
    dup
    rot
    mul
    swap
    load 1
    sub
    load NULL

check:
    dup
    load 1
    bgt loop
    drop
"""

# A broken version of SOURCE_BASIC which falls through to "check"
# with a pointer at the top of the stack instead of the expected
# integer that should be caught by the merge.
SOURCE_BAD_TYPE = """\
define test::factorial returns int
    argument int x

    load 1
    swap
    goto check

loop:
    dup
    rot
    mul
    swap
    load 1
    sub
    cast 0, ptr

check:
    dup
    load 1
    bgt loop
    drop
"""

# A broken version of SOURCE_BASIC that trys to use "load x" at
# the start of the loop.  The name "x" should have been removed by
# the merge as it does not exist when falling through to "check".
SOURCE_LOST_NAME = """\
define test::factorial returns int
    argument int x

    load 1
    swap
    goto check

loop:
    load x
    rot
    mul
    swap
    load 1
    sub

check:
    dup
    load 1
    bgt loop
    drop
"""

# A fixed version of SOURCE_LOST_NAME where the name "x" is restored
# at the end of the loop and so not removed by the merge.
SOURCE_GOT_NAME = """\
define test::factorial returns int
    argument int x

    load 1
    swap
    goto check

loop:
    load x
    rot
    mul
    swap
    load 1
    sub
    name 0, x

check:
    dup
    load 1
    bgt loop
    drop
"""

FACTORIALS = (
    (0, 1),
    (1, 1),
    (2, 2),
    (3, 6),
    (4, 24),
    (5, 120),
    (6, 720),
    (7, 5040),
    (8, 40320),
    (9, 362880),
    (10, 3628800),
    (11, 39916800),
    (12, 479001600))

class TestLoops(TestCase):
    def test_basic(self):
        """Check that loops work."""
        self.__test(SOURCE_BASIC)

    def test_bad_depth(self):
        """Check that merges with incorrect depth are caught."""
        self.__test(SOURCE_BAD_DEPTH, StackMergeError)

    def test_bad_type(self):
        """Check that merges with incorrect types are caught."""
        self.__test(SOURCE_BAD_TYPE, StackMergeError)

    def test_lost_name(self):
        """Check that merges with lost names are caught."""
        self.__test(SOURCE_LOST_NAME, UndefinedIdentError)

    def test_got_name(self):
        """Check that lost names can be restored."""
        self.__test(SOURCE_GOT_NAME)

    def __test(self, source, exception=None):
        if exception is None:
            tree, output = self.compile(source)
            sig = output.note.signature
            self.assertEqual(sig, "test::factorial(i)i")
            for input, expect in FACTORIALS:
                self.assertEqual(output.call(sig, input), [expect])
        else:
            self.assertRaises(exception, self.compile, source)
